import pysam
import Bio.Seq

import handygenome.refgenome as refgenome
import handygenome.interval as libinterval
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.sv.breakends as breakends
import handygenome.variant.vcfspec as libvcfspec


MITOCHONDRIAL = ('chrM', 'MT')


class SimpleStructuralVariant(libinterval.Interval):

	def __init__(self, chrom, *, start1=None, end1=None, start0=None, end0=None, fasta=None, refver=None):
		"""
		'chrom' is mandatory
		('start1' and 'end1') or ('start0' and 'end0') must be given(for coordinate setting).
		'fasta' or 'refver' may be given

		'start0' and 'end0' are 0-based half-open system.
		'start1' and 'end1' are 1-based closed system.
		The interval given by [start0, end0) or [start1, end1] itself represents the affected bases. 
			(e.g. The bases on 'start1' or 'end1' are affected)
		The sequence represented by the interval [start0, end0) is affected.
		"""

		super().__init__(chrom, start1, end1, start0, end0)

		if fasta is None:
			self.fasta = refgenome.get_default_fasta(refver)
		else:
			self.fasta = fasta

	def __repr__(self):
		return f'<{self.__class__.__name__}> {self.chrom}:{self.start1:,}-{self.end1:,}'

	def get_seq(self):
		return self.fasta.fetch(self.chrom, self.start0, self.end0)

	def get_vr_symbolic(self):
		chromdict = refgenome.ChromDict.from_fasta(self.fasta)
		vr = initvcf.create_vr(chromdict=chromdict)
		vr.contig = self.chrom
		vr.pos = self.start0
		vr.ref = self.fasta.fetch(self.chrom, vr.pos - 1, vr.pos)
		vr.alts = [self.__class__.ALTsymbol]
		vr.stop = self.end0

		return vr


class Deletion(SimpleStructuralVariant):

	ALTsymbol = '<DEL>'

	def get_hgvsg(self):
		prefix = 'm' if (self.chrom in MITOCHONDRIAL) else 'g'
		if self.length == 1:
			hgvsg = f'{self.chrom}:{prefix}.{self.start0 + 1}del'
		else:
			hgvsg = f'{self.chrom}:{prefix}.{self.start0 + 1}_{self.end0}del'

		return hgvsg

	def get_vcfspec(self):
		pos = self.start0
		ref = self.fasta.fetch(self.chrom, pos - 1, self.end0)
		alt = ref[0]

		return libvcfspec.Vcfspec(self.chrom, pos, ref, [alt])
			
	def get_bnds(self):
		return breakends.Breakends(
				chrom1 = self.chrom,
				pos1 = self.start0,
				pos1_endis5 = False,
				chrom2 = self.chrom,
				pos2 = self.end0 + 1,
				pos2_endis5 = True,
				inserted_seq = list(),
				fasta = self.fasta,
				)


class TandemDuplication(SimpleStructuralVariant):

	ALTsymbol = '<DUP:TANDEM>'

	def get_hgvsg(self):
		prefix = 'm' if (self.chrom in MITOCHONDRIAL) else 'g'
		if self.length == 1:
			hgvsg = f'{self.chrom}:{prefix}.{self.start0 + 1}dup'
		else:
			hgvsg = f'{self.chrom}:{prefix}.{self.start0 + 1}_{self.end0}dup'

		return hgvsg

	def get_vcfspec(self):
		pos = self.start0
		fetched = self.fasta.fetch(self.chrom, pos - 1, self.end0)
		ref = fetched[0]
		alt = fetched

		return libvcfspec.Vcfspec(self.chrom, pos, ref, [alt])


class Inversion(SimpleStructuralVariant):

	ALTsymbol = '<INV>'

	def get_hgvsg(self):
		prefix = 'm' if (self.chrom in MITOCHONDRIAL) else 'g'
		hgvsg = f'{self.chrom}:{prefix}.{self.start0 + 1}_{self.end0}inv'

		return hgvsg

	def get_vcfspec(self, padded = False):
		if padded:
			pos = self.start0
			fetched = self.fasta.fetch(self.chrom, pos - 1, self.end0)
			ref = fetched
			alt = fetched[0] + Bio.Seq.reverse_complement(fetched[1:])
		else:
			pos = self.start0 + 1
			fetched = self.fasta.fetch(self.chrom, pos - 1, self.end0)
			ref = fetched
			alt = Bio.Seq.reverse_complement(fetched)

		return libvcfspec.Vcfspec(self.chrom, pos, ref, [alt])




import re
import os
import argparse

import pysam

import handygenome.vcfeditor.misc as vcfmisc
import handygenome.workflow as workflow


def sanity_check_args(infile_path_list, outfile_path, outfile_must_not_exist):
    infile_path_list = [workflow.arghandler_infile(x) 
                        for x in infile_path_list]
    arghandler = workflow.get_arghandler_outfile(outfile_must_not_exist)
    outfile_path = arghandler(outfile_path)

    return infile_path_list, outfile_path


def sanity_check_headers(infile_path_list):
    """Checks if all headers are equal"""

    samples_list = list()
    infokeys_list = list()
    formatkeys_list = list()

    for vcf_path in infile_path_list:
        with pysam.VariantFile(vcf_path, 'r') as in_vcf:
            samples_list.append(tuple(in_vcf.header.samples))
            infokeys_list.append(tuple(sorted(in_vcf.header.info.keys())))
            formatkeys_list.append(tuple(sorted(in_vcf.header.formats.keys())))

    # check samples
    if len(set(samples_list)) != 1:
        raise Exception(f'Headers of input vcf files are different in '
                        'samples (orders must match)')
    elif len(set(infokeys_list)) != 1:
        raise Exception(f'Headers of input vcf files are different in '
                        'INFO keys')
    elif len(set(formatkeys_list)) != 1:
        raise Exception(f'Headers of input vcf files are different in '
                        'FORMAT keys')


def write_outfile(infile_path_list, outfile_path, mode_pysam):
    with pysam.VariantFile(infile_path_list[0]) as in_vcf:
        outfile_header = in_vcf.header.copy()

    with pysam.VariantFile(outfile_path, mode=mode_pysam, 
                           header=outfile_header) as out_vcf:
        for vcf_path in infile_path_list:
            with pysam.VariantFile(vcf_path, 'r') as in_vcf:
                for vr in in_vcf.fetch():
                    out_vcf.write(vr)


def main(
        infile_path_list,
        outfile_path,
        mode_bcftools='z', 
        mode_pysam=None, 
        outfile_must_not_exist='ask',
        ):
    infile_path_list, outfile_path = sanity_check_args(
        infile_path_list, outfile_path, outfile_must_not_exist)
    sanity_check_headers(infile_path_list)

    mode_pysam = vcfmisc.write_mode_arghandler(mode_bcftools, mode_pysam)
    write_outfile(infile_path_list, outfile_path, mode_pysam)



#!/usr/bin/env python3

"""
Created: March 10th, 2020
@author: Mark Barnes

PlotDecomposition works with matrix formats SigProfiler SBS-96, SBS-1536, DBS-78,
and ID-83. This program is intended to take two matrices.

(1) Sample matrix - A SigProfiler formatted SBS-96, SBS-1536, DBS-78, or ID-83
matrix.
(2) Basis matrix - A SigProfiler formatted SBS-96, SBS-1536, DBS-78, or ID-83
matrix that is the decomposition of (1).

When running the function 'run_PlotDecomposition' a plot of the decomposition will
be generated and saved to the output folder. Refer to the function below to learn
more about the parameters required to generate the decomposition plot.
"""

import os
import pandas as pd
import numpy as np
import scipy.stats
import sigProfilerPlotting as sigPlt
import shutil
import SigProfilerAssignment
import SigProfilerAssignment.DecompositionPlots
from SigProfilerAssignment.DecompositionPlots import SigProfilerPlottingMatrix as mPlt
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_SBS96 as spd_96
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_SBS288 as spd_288
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_SBS1536 as spd_1536
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_DBS78 as spd_78
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_ID83 as spd_83
from SigProfilerAssignment.DecompositionPlots import PlotDecomposition_CNV48 as cnv_48
from SigProfilerAssignment import decompose_subroutines as sub
# imports for working with plots in memory
import io
from PIL import Image
from reportlab.lib.utils import ImageReader
import json
import base64
# Global Variables
SBS_CONTEXTS = ["6", "24", "96", "288", "384", "1536", "6144"]
DBS_CONTEXTS = ["78", "186", "1248", "2976"]
ID_CONTEXTS = ["28", "83", "415"]
CNV_CONTEXTS = ["48"]
MTYPE_OPTIONS = ["6", "24", "96", "384", "1536", "6144", "28", "83", "415", "78", "186", "1248", "2976"]
DECOMPOSITION_PATH = SigProfilerAssignment.DecompositionPlots.__path__[0]
REFSIG_PATH = os.path.join(SigProfilerAssignment.__path__[0], "data/Reference_Signatures")
TEMPLATE_PATH = os.path.join(DECOMPOSITION_PATH,'CosmicTemplates')

# Remove templates so that they can be rebuilt
def remove_cosmic_templates():
	if not os.path.exists(TEMPLATE_PATH):
		print("No files are installed at: ", TEMPLATE_PATH)
	try:
		shutil.rmtree(TEMPLATE_PATH)
	except OSError as e:
		print("Error: %s : %s" % (TEMPLATE_PATH, e.strerror))

# Create a set of serialized JSON reference signature plots for fast loading
def install_cosmic_plots(context_type="96", genome_build="GRCh37", cosmic_version="3.3", exome=False):

	if not os.path.exists(TEMPLATE_PATH):
		os.mkdir(TEMPLATE_PATH)

	# determine if context is from SBS, ID, DBS, or CNV
	context_type_str = ""
	if context_type in SBS_CONTEXTS:
		context_type_str="SBS"
	elif context_type in DBS_CONTEXTS:
		context_type_str="DBS"
	elif context_type in ID_CONTEXTS:
		context_type_str="ID"
	elif context_type in CNV_CONTEXTS:
		context_type_str="CNV"
	else:
		raise ValueError("ERROR: context", context_type, "not in context lists.")

	# if exome is true, append _exome to end of file name
	exome_str = ""
	if exome:
		exome_str = "_exome"

	cosmic_file_name = "COSMIC_v" + str(cosmic_version) + "_" + \
						context_type_str + "_" + genome_build + \
						exome_str + ".txt"
	json_file_name = "COSMIC_v" + str(cosmic_version) + "_" + \
						context_type_str + "_" + genome_build + \
						exome_str + ".json"

	# Load cosmic plots if they exist
	filename= os.path.join(TEMPLATE_PATH, json_file_name)
	if os.path.exists(filename):
		cosmic_buff_bytes = json.load(open(filename))
		cosmic_buff_plots = {}
		# Read from JSON, decode, and convert to bytesIO
		for tmp_plots in cosmic_buff_bytes.keys():
			cosmic_buff_plots[tmp_plots] = io.BytesIO(base64.b64decode(cosmic_buff_bytes[tmp_plots]))
		return cosmic_buff_plots

	# Generate cosmic plots if they were not found
	else:
		cosmic_file_path = os.path.join(REFSIG_PATH, genome_build, cosmic_file_name)
		json_file_path = os.path.join(TEMPLATE_PATH, json_file_name)
		print("Generating plots for",  "COSMIC_v" + str(cosmic_version) + "_" + \
						context_type_str + "_" + genome_build + \
						exome_str,"now...")

		# Create the respective plots
		if context_type_str == "SBS":
			cosmic_buff_plots = sigPlt.plotSBS(cosmic_file_path, "buffer",
							"buffer", context_type, percentage=True,
							savefig_format="buffer_stream")
		elif context_type_str == "DBS":
			cosmic_mtx = pd.read_csv(cosmic_file_path, sep="\t")
			cosmic_buff_plots = mPlt.plotDBS(cosmic_mtx, "buffer",
						"buffer", context_type, percentage=True)
		elif context_type_str == "ID":
			cosmic_buff_plots = sigPlt.plotID(cosmic_file_path, "buffer",
						"buffer", context_type, percentage=True,
						savefig_format="buffer_stream")

		# Process the plots to be stored in JSON file
		cosmic_img_dict = {}
		for tmp_plot in cosmic_buff_plots.keys():
			# start at beggining of binary file
			seek_start = cosmic_buff_plots[tmp_plot].seek(0)
			encoded = base64.b64encode(cosmic_buff_plots[tmp_plot].getvalue())
			cosmic_img_dict[tmp_plot] = encoded.decode('ascii')

		# JSON output processing
		json_object = json.dumps(cosmic_img_dict)
		with open(json_file_path, "w") as outfile:
			outfile.write(json_object)

		print("Plots for", "COSMIC_v" + str(cosmic_version) + "_" + \
						context_type_str + "_" + genome_build + \
						exome_str, "have been successfully installed.")
		return cosmic_buff_plots


# Helper function for converting BytesIO to image so it can be plotted by reportlab
def bytes_to_img(byte_png):
	byte_png.seek(0)
	tmp_im=Image.open(byte_png)
	image = ImageReader(tmp_im)
	return image
	
# Helper function to convert byte array to image array
def open_byte_to_img_dict(byte_dict):

	img_dict = dict()
	for name in byte_dict.keys():
		tmp_img = bytes_to_img(byte_dict[name])
		img_dict[name] = tmp_img
	return img_dict

def calculate_similarities(denovo, denovo_name, est_denovo):
	from numpy import inf

	# If matrix is 1536 context, then collapse it to 96 format
	if denovo.shape[0]==1536:
		index = denovo.iloc[:,0]
		denovo_tmp = pd.DataFrame(denovo, index=index)
		denovo_tmp = denovo.groupby(denovo_tmp.index.str[1:8]).sum()
		denovo = pd.DataFrame(denovo_tmp)
		denovo = denovo.reset_index()
	elif denovo.shape[0]==288:
		index = denovo.iloc[:,0]
		denovo_tmp = pd.DataFrame(denovo, index=index)
		denovo_tmp = denovo.groupby(denovo_tmp.index.str[2:9]).sum()
		denovo = pd.DataFrame(denovo_tmp)
		denovo = denovo.reset_index()


	sample_names = [denovo_name]

	if  sample_names is False:
		sample_names = ["None"]*denovo.shape[1]

	cosine_similarity_list = []
	cosine_distance_list = []
	correlation_list = []
	correlation_distance_list = []
	kl_divergence_list = []
	l1_norm_list = []
	l2_norm_list = []
	relative_l1_list = []
	relative_l2_list = []

	p_i = denovo[denovo_name]
	q_i = est_denovo
	
	cosine_similarity_list.append(round(sub.cos_sim(p_i,q_i ),3))
	cosine_distance_list.append(round(scipy.spatial.distance.cosine(p_i, q_i),3))
	correlation_list.append(round(scipy.stats.pearsonr(p_i,q_i)[0],3))
	correlation_distance_list.append(round(1-scipy.stats.pearsonr(p_i,q_i)[0],3))
	kl_divergence_list.append(round(scipy.stats.entropy(p_i,q_i),4))
	l1_norm_list.append(round(np.linalg.norm(p_i-q_i , ord=1),2))
	relative_l1_list.append(round((l1_norm_list[-1]/np.linalg.norm(p_i, ord=1))*100,3))
	l2_norm_list.append(round(np.linalg.norm(p_i-q_i , ord=2),2))
	relative_l2_list.append(round((l2_norm_list[-1]/np.linalg.norm(p_i, ord=2))*100,3))
	kl_divergence_list = np.array(kl_divergence_list)
	kl_divergence_list[kl_divergence_list == inf] =1000
	
	similarities_dataframe = pd.DataFrame({"Sample Names": sample_names, \
										   "Cosine Similarity": cosine_similarity_list, \
										   "Cosine Distance": cosine_distance_list, \
										   "Correlation Distance": correlation_distance_list, \
										   "Correlation Coefficient": correlation_list, \
										   "L1 Norm": l1_norm_list, \
										   "L1 Norm %":relative_l1_list, \
										   "L2 Norm": l2_norm_list, \
										   "L2 Norm %": relative_l2_list, \
										   "KL Divergence": kl_divergence_list})
	similarities_dataframe = similarities_dataframe.set_index("Sample Names")

	return similarities_dataframe


def genSBS_pngs(denovo_mtx, basis_mtx, output_path, project, mtype, ss_decomp=False):
	denovo_plots = dict()
	basis_plots = dict()
	if mtype == "1536" or mtype == "288":
		denovo_plots = mPlt.plotSBS(denovo_mtx, output_path, project, mtype,
					percentage=True)
		if basis_mtx is not None:
			basis_plots = sigPlt.plotSBS(basis_mtx, output_path, project, "96",
					percentage=True, savefig_format="buffer_stream")
	elif mtype == "96":
		denovo_plots = sigPlt.plotSBS(denovo_mtx, output_path, project, mtype,
			percentage=(not ss_decomp), savefig_format="buffer_stream")
		if basis_mtx is not None:
			basis_plots = sigPlt.plotSBS(basis_mtx, output_path, project, mtype,
					percentage=True, savefig_format="buffer_stream")
	return denovo_plots,basis_plots

def genDBS_pngs(denovo_mtx, basis_mtx, output_path, project, mtype):
	denovo_plots = dict()
	basis_plots = dict()
	denovo_plots = mPlt.plotDBS(denovo_mtx, output_path, project, mtype,
			percentage=True)
	if basis_mtx is not None:
		basis_plots = mPlt.plotDBS(basis_mtx, output_path, project, mtype,
				percentage=True)
	return denovo_plots,basis_plots

def genID_pngs(denovo_mtx, basis_mtx, output_path, project, mtype):
	denovo_plots = dict()
	basis_plots = dict()
	denovo_plots = sigPlt.plotID(denovo_mtx, output_path, project, mtype,
			percentage=True, savefig_format="buffer_stream")
	if basis_mtx is not None:
		basis_plots = sigPlt.plotID(basis_mtx, output_path, project, mtype,
				percentage=True, savefig_format="buffer_stream")
	return denovo_plots,basis_plots
	
def genCNV_pngs(denovo_mtx, basis_mtx, output_path, project, mtype):
	denovo_plots = dict()
	basis_plots = dict()
	denovo_plots = sigPlt.plotCNV(denovo_mtx, output_path, project,
				plot_type="pdf", percentage=True, aggregate=False,
				read_from_file=False, write_to_file=False)

	if basis_mtx is not None:
		basis_plots = sigPlt.plotCNV(basis_mtx, output_path, project,
				plot_type="pdf", percentage=True, aggregate=False,
				read_from_file=False, write_to_file=False)
	return denovo_plots,basis_plots

# signames, weights
def gen_sub_plots(denovo_mtx, basis_mtx, output_path, project, mtype, ss_decomp):

	# Make output directory
	if not os.path.exists(output_path):
		os.makedirs(output_path)

	if mtype in SBS_CONTEXTS:
		denovo_plots,basis_plots = genSBS_pngs(denovo_mtx, basis_mtx,
									output_path, project, mtype, ss_decomp)
		return denovo_plots,basis_plots

	elif mtype in DBS_CONTEXTS:
		denovo_plots,basis_plots = genDBS_pngs(denovo_mtx, basis_mtx,
									output_path, project, mtype)
		return denovo_plots,basis_plots

	elif mtype in ID_CONTEXTS:
		denovo_plots,basis_plots=genID_pngs(denovo_mtx, basis_mtx,
									output_path, project, mtype)
		return denovo_plots,basis_plots

	elif mtype in CNV_CONTEXTS:
		denovo_plots, basis_plots=genCNV_pngs(denovo_mtx, basis_mtx,
									output_path, project, mtype)
		return denovo_plots,basis_plots

	else:
		print("ERROR: mtype is " + mtype + " and is not yet supported.")


# generate the plot for the reconstruction
def gen_reconstructed_png_percent(denovo_name, basis_mtx, basis_names, weights, output_path, project, mtype):
	reconstruction_plot=dict()
	mut_col = basis_mtx.iloc[:,0]

	recon_plot = basis_mtx[basis_names[0]]*float(weights[0].strip("%"))/100

	for i in range(1,len(weights)):
		recon_plot = recon_plot + basis_mtx[basis_names[i]]*(float(weights[i].strip("%"))/100)

	recon_plot = pd.Series(recon_plot, name=denovo_name)
	reconstruction_mtx = pd.concat([mut_col, recon_plot], axis=1)
	if mtype in SBS_CONTEXTS:
		if mtype == "1536" or mtype == "288":
			reconstruction_plot=sigPlt.plotSBS(reconstruction_mtx, output_path, \
				"reconstruction_" + project, "96", percentage=True, savefig_format='buffer_stream')
		else:
			reconstruction_plot=sigPlt.plotSBS(reconstruction_mtx, output_path, \
				"reconstruction_" + project, mtype, percentage=True, savefig_format='buffer_stream')
	elif mtype in DBS_CONTEXTS:
		reconstruction_plot=mPlt.plotDBS(reconstruction_mtx, output_path,
				"reconstruction_" + project, mtype, percentage=True)
	elif mtype in ID_CONTEXTS:
		reconstruction_plot=sigPlt.plotID(reconstruction_mtx, output_path,
				"reconstruction_" + project, mtype, percentage=True, savefig_format='buffer_stream')
	elif mtype in CNV_CONTEXTS:
		reconstruction_plot = sigPlt.plotCNV(reconstruction_mtx, output_path, "reconstruction_"+project, plot_type="pdf", \
										percentage=True, aggregate=False, read_from_file=False, write_to_file=False)
	else:
		print("ERROR: mtype is " + mtype + " and is not yet supported.")

	return reconstruction_mtx, reconstruction_plot

# generate the plot for the reconstruction
def gen_reconstructed_png_numerical(denovo_mtx, denovo_name, basis_mtx, basis_names, weights, output_path, project, mtype):
	sample_tmb = denovo_mtx[denovo_name].sum()
	reconstruction_plot=dict()
	mut_col = basis_mtx.iloc[:,0]

	recon_plot = basis_mtx[basis_names[0]]*float(weights[0].strip("%"))/100*sample_tmb
	for i in range(1,len(weights)):
		recon_plot = recon_plot + basis_mtx[basis_names[i]]*(float(weights[i].strip("%"))/100)*sample_tmb

	recon_plot = pd.Series(recon_plot, name=denovo_name)
	reconstruction_mtx = pd.concat([mut_col, recon_plot], axis=1)
	reconstruction_mtx = reconstruction_mtx.round(0)
	if mtype in SBS_CONTEXTS:
		if mtype == "1536" or mtype == "288":
			reconstruction_plot=sigPlt.plotSBS(reconstruction_mtx, output_path,
						"reconstruction_" + project, "96", percentage=False)
		else:
			reconstruction_plot=sigPlt.plotSBS(reconstruction_mtx, output_path,
						"reconstruction_" + project, mtype, percentage=False,
						savefig_format="buffer_stream")
	elif mtype in DBS_CONTEXTS:
		reconstruction_plot=mPlt.plotDBS(reconstruction_mtx, output_path,
				"reconstruction_" + project, mtype, percentage=False)
	elif mtype in ID_CONTEXTS:
		reconstruction_plot=sigPlt.plotID(reconstruction_mtx, output_path,
				"reconstruction_" + project, mtype, percentage=False)
	elif mtype in CNV_CONTEXTS:
		reconstruction_plot = sigPlt.plotCNV(reconstruction_mtx, output_path,
				"reconstruction_"+project, plot_type="pdf", percentage=True,
				aggregate=False, read_from_file=False, write_to_file=False)
	else:
		print("ERROR: mtype is " + mtype + " and is not yet supported.")

	return reconstruction_mtx, reconstruction_plot


def gen_decomposition(denovo_name, basis_names, weights, output_path, project, \
	mtype, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict, \
	reconstruction=False, statistics=None, cosmic_version=None, custom_text=None):

	"""
	Generate the correct plot based on mtype.

	Parameters:
	----------
	denovo_name: 				(String) 			Name of denovo signature
	basis_names: 				(List of Strings) 	Names of basis signatures
	weights:					(List of Strings) 	Percentile contribution for each basis signature
	output_path: 				(String) 			Path to existing output directory
	project: 					(String) 			Project name appended to file names
	mtype: 						(String) 			The context 'mtype_options' has valid values
	denovo_plots_dict			(Dictionary)		Signatures are keys, ByteIO plots are values
	basis_plots_dict			(Dictionary)		Signatures are keys, ByteIO plots are values
	reconstruction_plot_dict	(Dictionary)		Signatures are keys, ByteIO plots are values
	reconstruction: 			(Boolean) 			True to generate plot w/ reconstruction
	statistics: 				(Pandas Dataframe) 	Output from calculate_similarities()
	"""

	if mtype == "6":
		print("Need to add support for SBS6 Decomposition")
	elif mtype == "24":
		print("Need to add support for SBS24 Decomposition")
	elif mtype == "96":
		byte_plot=spd_96.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot
	elif mtype == "288":
		byte_plot=spd_288.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot
	elif mtype == "384":
		print("Need to add support for SBS24 Decomposition")
	elif mtype == "1536":
		byte_plot=spd_1536.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot
	elif mtype == "6144":
		print("Need to add support for SBS6144 Decomposition")
	elif mtype == "28":
		print("Need to add support for ID28 Decomposition")
	elif mtype == "83":
		byte_plot=spd_83.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot
	elif mtype == "415":
		print("Need to add support for ID415 Decomposition")
	elif mtype == "78":
		byte_plot=spd_78.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot
	elif mtype == "186":
		print("Need to add support for DBS186 Decomposition")
	elif mtype == "1248":
		print("Need to add support for DBS1248 Decomposition")
	elif mtype == "2976":
		print("Need to add support for DBS2976 Decomposition")
	elif mtype == "48":
		byte_plot=cnv_48.gen_decomposition(denovo_name, basis_names, weights, output_path,
			project, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction, statistics, cosmic_version, custom_text)
		return byte_plot



def run_PlotDecomposition(denovo_mtx, denovo_name, basis_mtx, basis_names,
		weights, nonzero_exposures, output_path, project, mtype,
		cosmic_version="3.3", genome_build="GRCh37", exome=False,
		custom_text=None):
	"""
	Generates a decomposition plot of the denovo_mtx using the basis_mtx.

	Parameters:
	----------

	denovo_mtx: Pandas Dataframe. This format represents the catalog of mutations seperated by tab.

	denovo_name: String. The name of the one sample in denovo_mtx to decompose.

	basis_mtx: Pandas Dataframe. This format represents the catalog of mutations seperated by tab.

	basis_names: List of Strings. The names of the samples in denovo_mtx that
	the denovo_name sample from denovo_mtx is decomposed into.
	ie. basis_names=["SBS1", "SBS5", "SBS15", "SBS20"]

	weights: List of Strings. The percentile weight corresponding to each basis
	in basis_names. Refer to example function call below for more detail.
	ie. weights=["11.58%", "42.38%", "16.46%", "29.58%"]

	output_path: String. Path to where to store the output.

	project: String. This string is appended to the file name output.

	mtype: String. The context of the data. Valid values include: "96", "1536","78", and "83".

	cosmic_version: String. The version of signatures being used.

	custom_text: String. A custom message displayed on decomposition plot.

	Returns:
	-------
	None.
	"""

	# Create the denovo plots and load basis plots
	if mtype == "48":
		denovo_plots_dict, basis_plots_dict = gen_sub_plots(denovo_mtx, basis_mtx,
				output_path, project, mtype, ss_decomp=False)
	else:
		denovo_plots_dict = gen_sub_plots(denovo_mtx, None,
				output_path, project, mtype, ss_decomp=False)
		denovo_plots_dict = denovo_plots_dict[0]
	# Create the matrix and plot for the reconstructed matrix
	reconstructed_mtx,reconstruction_plot_dict = gen_reconstructed_png_percent(
		denovo_name, basis_mtx, basis_names, weights, output_path, project, mtype)
	# Create a subset matrix with the present signatures
	present_sigs=np.array(basis_mtx[basis_names])
	reconstructed_mtx = np.dot(present_sigs,nonzero_exposures)
	# Convert dictionary of bytes to dictionary of images
	denovo_plots_dict = open_byte_to_img_dict(denovo_plots_dict)
	# Load in the COSMIC plots
	if mtype != "48":
		basis_plots_dict = install_cosmic_plots(context_type=mtype,
				genome_build=genome_build, cosmic_version=cosmic_version,
				exome=exome)
		basis_plots_dict = {key: basis_plots_dict[key] for key in basis_names}
	basis_plots_dict = open_byte_to_img_dict(basis_plots_dict)
	# Generate the reconstruction plot
	reconstruction_plot_dict = open_byte_to_img_dict(reconstruction_plot_dict)
	# Get the reconstruction statistics
	statistics=calculate_similarities(denovo_mtx, denovo_name, reconstructed_mtx)
	# Return the decomposition plot as a byte array
	byte_plot = gen_decomposition(denovo_name, basis_names, weights, output_path, project,
		mtype, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
		reconstruction=True, statistics=statistics, cosmic_version=cosmic_version,
		custom_text=custom_text)
	
	return byte_plot

# context="96", genome_build="GRCh37", cosmic_version="3.3", exome=False
def run_PlotSSDecomposition(denovo_mtx, denovo_name, basis_mtx, basis_names, \
		weights, output_path, project, context_type, genome_build="GRCh37", \
		cosmic_version="3.3", custom_text=None, exome=False):
	"""
	Generates a reconstruction of a sample given a set of signatures.

	Parameters:
	----------

	samples: Pandas Dataframe. Samples and their channel counts.
			df_allgenomes =  pd.DataFrame(allgenomes, columns=allcolnames)

	denovo_mtx: Pandas Dataframe. This format represents the catalog of mutations seperated by tab.

	denovo_name: String. The name of the one sample in denovo_mtx to decompose.

	basis_mtx: Pandas Dataframe. This format represents the catalog of mutations seperated by tab.

	basis_names: List of Strings. The names of the samples in denovo_mtx that
	the denovo_name sample from denovo_mtx is decomposed into.
	ie. basis_names=["SBS1", "SBS5", "SBS15", "SBS20"]

	weights: List of Strings. The percentile weight corresponding to each basis
	in basis_names. Refer to example function call below for more detail.
	ie. weights=["11.58%", "42.38%", "16.46%", "29.58%"]

	output_path: String. Path to where to store the output.

	project: String. This string is appended to the file name output.

	context_type: String. The context of the data. Valid values include: "96", "1536","78", and "83".

	genome_build: String. The genome being used.

	cosmic_version: String. The version of signatures being used.

	custom_text: String. A custom message displayed on decomposition plot.

	exome: Boolean. True if using exome COSMIC signatures, and False if not.

	Returns:
	-------
	None.
	"""

	# Create the denovo plots
	denovo_plots_dict=gen_sub_plots(denovo_mtx, None, output_path,
			project, context_type, ss_decomp=True)
	denovo_plots_dict = denovo_plots_dict[0]
	# Load in the COSMIC plots
	basis_plots_dict = install_cosmic_plots(context_type=context_type,
			genome_build=genome_build, cosmic_version=cosmic_version,
			exome=exome)

	# Create reconstructed matrix and plot
	reconstructed_mtx,reconstruction_plot_dict = gen_reconstructed_png_numerical(
		denovo_mtx, denovo_name, basis_mtx,
		basis_names, weights, output_path, project, context_type)

	denovo_plots_dict = open_byte_to_img_dict(denovo_plots_dict)
	# subset basis_plots_dict to only the plots used
	basis_plots_dict = {key: basis_plots_dict[key] for key in basis_names}
	basis_plots_dict = open_byte_to_img_dict(basis_plots_dict)

	reconstruction_plot_dict = open_byte_to_img_dict(reconstruction_plot_dict)

	statistics=calculate_similarities(denovo_mtx, denovo_name, reconstructed_mtx[denovo_name])
	byte_plot = gen_decomposition(denovo_name, basis_names, weights, output_path, project,
			context_type, denovo_plots_dict, basis_plots_dict, reconstruction_plot_dict,
			reconstruction=True, statistics=statistics, cosmic_version=cosmic_version,
			custom_text=custom_text)


	return byte_plot


import sys
import collections
import itertools
import functools
import logging
import inspect
import random
import uuid

import pysam
import Bio.Align
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.deco as deco
import handygenome.workflow as workflow
import handygenome.variant.vcfspec as libvcfspec
import handygenome.read.pileup as libpileup
import handygenome.align.alignhandler as alignhandler
import handygenome.bameditor as bameditor
import handygenome.read.readhandler as readhandler
import handygenome.read.readplus as readplus
import handygenome.align.realign.base as realign_base


class RealignerPileup(realign_base.RealignerPileupBase):
    # initializers
    def __init__(
        self, 
        bam, 
        chrom, 
        start0=None, 
        end0=None, 
        refver=None, 
        fasta=None,

        aligner=realign_base.DEFAULT_ALIGNER,
        init_df=True,

        verbose=False,
        logger=None,

        **kwargs,
    ):
        """Args:
            start0 and end0 are only required when init_df == True
        """
        # initiation
        libpileup.PileupBase.__init__(
            self, 
            bam=bam, 
            chrom=chrom, start0=start0, end0=end0, 
            refver=refver, fasta=fasta,
            init_df=init_df, 
            verbose=verbose, logger=logger,
        )

        # set other parameters
        self.params = realign_base.parse_rpileup_kwargs(**kwargs)
        self.aligner = aligner
        if init_df:
            self._set_active_info()

        self.hit_left_margin = False
        self.hit_right_margin = False

    @classmethod
    def init_and_augment(
        cls, 
        bam, 
        chrom, 
        start0, 
        end0, 
        refver=None, 
        fasta=None,
        aligner=realign_base.DEFAULT_ALIGNER,
        verbose=False, 
        logger=None, 
        **kwargs,
    ):
        result = cls(
            bam=bam, 
            chrom=chrom, start0=start0, end0=end0, 
            refver=refver, fasta=fasta,
            aligner=aligner,

            init_df=True,

            verbose=verbose,
            logger=logger,

            **kwargs,
        )
        secresult_inactive, secresult_vcfspec = result.augment_margins()

        return result, secresult_inactive, secresult_vcfspec

    def spawn(self):
        result = self.__class__(
            bam=self.bam,
            chrom=self.chrom,
            refver=self.refver,
            fasta=self.fasta,
            aligner=self.aligner,
            init_df=False,
            verbose=self.verbose,
            logger=self.logger,
            **self.params,
        )
        result.read_store = self.read_store
        return result
    
    def subset(self, start0, end0, inplace=False):
        new_active_info = self.active_info.loc[start0:(end0 - 1)]
        if inplace:
            libpileup.PileupBase._subset_base(self, start0, end0, inplace=inplace)
            self.active_info = new_active_info
        else:
            result = libpileup.PileupBase._subset_base(self, start0, end0, inplace=inplace)
            result.active_info = new_active_info
            return result

    def merge(self, other, other_on_left):
        self._merge_base(other, other_on_left=other_on_left)
        if other_on_left:
            self.active_info = pd.concat([other.active_info, self.active_info])
        else:
            self.active_info = pd.concat([self.active_info, other.active_info])

    def split(self, start0_list):
        split_pileups = self._split_base(start0_list)
        for pileup in split_pileups:
            pileup.params = self.params.copy()
            pileup.prepare_vcfspecs()
        return split_pileups

    ################
    def check_touches_left_limit(self, start0=None):
        if start0 is None:
            start0 = self.start0
        return start0 <= self.params['start0_limit']

    def check_touches_right_limit(self, end0=None):
        if end0 is None:
            end0 = self.end0
        return end0 >= self.params['end0_limit']

    def check_touches_width_limit(self, start0=None, end0=None):
        if start0 is None:
            start0 = self.start0
        if end0 is None:
            end0 = self.end0
        return end0 - start0 >= self.params['max_pileup_width']

    def check_left_MQ_low(self):
        return self.MQ.iloc[0] < self.params['MQ_limit']

    def check_right_MQ_low(self):
        return self.MQ.iloc[-1] < self.params['MQ_limit']

    def check_left_depth_low(self):
        return self.get_depth(self.start0) < self.params['depth_limit']

    def check_right_depth_low(self):
        return self.get_depth(self.end0 - 1) < self.params['depth_limit']

    # extend and its helpers
    def extend_left(self, width):
        #assert self.start0 - width >= self.start0_limit
        self._extend_base(width, left=True)

    def extend_right(self, width):
        #assert self.end0 + width <= self.end0_limit
        self._extend_base(width, left=False)

    def _make_extend_pileup(self, width, left):
        if left:
            start0 = self.start0 - width
            end0 = self.start0
        else:
            start0 = self.end0
            end0 = self.end0 + width

        return self.__class__(
            bam=self.bam, 
            chrom=self.chrom, 
            start0=start0,
            end0=end0,
            refver=self.refver,
            fasta=self.fasta,
            aligner=self.aligner,
            init_df=True,
            verbose=self.verbose,
            logger=self.logger,
            **self.params,
        )

    def get_extend_generator(self):
        return realign_base.PileupExtendGenerator(self)

    # active info related ones #

    def get_inactive_length_left(self):
        for idx in range(self.width):
            if self.active_info.iloc[idx]:
                break
        if idx == self.width - 1:
            idx += 1
        return idx

    def get_inactive_length_right(self):
        for idx in range(-1, -1 -self.width, -1):
            if self.active_info.iloc[idx]:
                break
        if idx == -self.width:
            idx -= 1
        return -idx - 1

    # extension and reduction in search for realignment area
    def augment_margins(self):
        self.logger.debug(f'Beginning "augment_margins"')

        # immediately after seeding
        if self.active_info.any():
            self.logger.debug(f'\tskipping initial search')
        else:
            self.logger.debug(f'\tbeginning initial search')

            self.search_for_active_position()

            self.logger.debug(f'\tfinished initial search')

            if not self.active_info.any():
                self.logger.debug(f'Finished "augment_margins" without secure_margins')

                result_inactive = None
                result_vcfspec = None
                return result_inactive, result_vcfspec

        # trim excessive inactive padding
        self.reduce_left()
        self.reduce_right()
        # do after-discovery
        self.logger.debug(f'\tbeginning after-discovery')
        result_inactive, result_vcfspec = self.secure_margins()
        self.logger.debug(f'\tfinished after-discovery')

        self.logger.debug(f'Finished "augment_margins"')
        return result_inactive, result_vcfspec

    def search_for_active_position(self):
        gen = self.get_extend_generator()
        if gen.aborted or self.active_info.any():
            return
        
        while True:
            if gen.aborted:
                break

            try:
                gen.iter_left()
            except StopIteration:
                pass
            else:
                if gen.left_is_active:
                    break

            try:
                gen.iter_right()
            except StopIteration:
                pass
            else:
                if gen.right_is_active:
                    break

        # subset
        self.subset(gen.current_start0, gen.current_end0, inplace=True)

    def secure_margins(self):
        self.logger.debug(f'Began secure_margins')

        while True:
            self.logger.debug(f'Began secure_inactive_padding')
            result_inactive = self.secure_inactive_padding()
            self.logger.debug(f'Finished secure_inactive_padding. Margins: {self.range0}')

            self.logger.debug(f'Began prepare_vcfspecs')
            self.prepare_vcfspecs()
            self.logger.debug(f'Finished prepare_vcfspecs')

            self.logger.debug(f'Began secure_vcfspec_margins')
            result_vcfspec = self.secure_vcfspec_margins()
            self.logger.debug(f'Finished secure_vcfspec_margins. Margins: {self.range0}')

            if not result_vcfspec.edited:
                break
            #if (not result_inactive.left_okay) and (not result_inactive.right_okay):
            #    break
            #else:

        self.logger.debug(f'Finished secure_margins')

        return result_inactive, result_vcfspec

    # extension reduction helpers
    def reduce_left(self):
        if self.width > self.params['inactive_padding']:
            inactive_margin_length = self.get_inactive_length_left()
            if inactive_margin_length == self.params['inactive_padding']:
                left_fixed = True
            elif inactive_margin_length > self.params['inactive_padding']:
                new_start0 = self.start0 + (inactive_margin_length - self.params['inactive_padding'])
                self.subset(new_start0, self.end0, inplace=True)
                left_fixed = True
            else:
                left_fixed = False
        else:
            left_fixed = False

        return left_fixed

    def reduce_right(self):
        if len(self.range0) > self.params['inactive_padding']:
            inactive_margin_length = self.get_inactive_length_right()
            if inactive_margin_length == self.params['inactive_padding']:
                right_fixed = True
            elif inactive_margin_length > self.params['inactive_padding']:
                new_end0 = self.end0 - (inactive_margin_length - self.params['inactive_padding'])
                self.subset(self.start0, new_end0, inplace=True)
                right_fixed = True
            else:
                right_fixed = False
        else:
            right_fixed = False

        return right_fixed

    @classmethod
    def _securehelper_check_aborted(cls, gen, left_okay, right_okay):
        return (
            (
                cls._securehelper_check_left_blocked(gen, left_okay) and
                cls._securehelper_check_right_blocked(gen, right_okay)
            ) or
            gen.touched_width_limit
        )

    @classmethod
    def _securehelper_check_left_blocked(cls, gen, left_okay):
        return gen.left_is_blocked or gen.touched_width_limit or left_okay

    @classmethod
    def _securehelper_check_right_blocked(cls, gen, right_okay):
        return gen.right_is_blocked or gen.touched_width_limit or right_okay

    def secure_inactive_padding(self):
        # set params
        inactive_length_left = self.get_inactive_length_left()
        inactive_length_right = self.get_inactive_length_right()
        left_okay = (inactive_length_left >= self.params['inactive_padding'])
        right_okay = (inactive_length_right >= self.params['inactive_padding'])
        
        # initial abort check
        if left_okay and right_okay:
            return realign_base.SecureResult(
                touched_left_limit=self.check_touches_left_limit(),
                touched_right_limit=self.check_touches_right_limit(), 
                touched_width_limit=self.check_touches_width_limit(), 

                left_low_depth=self.check_left_depth_low(),
                right_low_depth=self.check_right_depth_low(),
                left_low_MQ=self.check_left_MQ_low(),
                right_low_MQ=self.check_right_MQ_low(),

                left_okay=left_okay, 
                right_okay=right_okay,
                edited=False,
            )

        # begin looping
        initial_start0 = self.start0
        initial_end0 = self.end0
        gen = self.get_extend_generator()
        while True:
            if self._securehelper_check_aborted(gen, left_okay, right_okay):
                break

            if not self._securehelper_check_left_blocked(gen, left_okay):
                gen.iter_left()
                if gen.left_is_active:
                    inactive_length_left = 0
                elif not gen.left_is_active:
                    inactive_length_left += 1

                left_okay = (inactive_length_left >= self.params['inactive_padding'])

            if not self._securehelper_check_right_blocked(gen, right_okay):
                gen.iter_right()
                if gen.right_is_active:
                    inactive_length_right = 0
                elif not gen.right_is_active:
                    inactive_length_right += 1

                right_okay = (inactive_length_right >= self.params['inactive_padding'])

        if (initial_start0 != gen.current_start0) or (initial_end0 != gen.current_end0):
            edited = True
            self.subset(gen.current_start0, gen.current_end0, inplace=True)
        else:
            edited = False

        return realign_base.SecureResult(
            touched_left_limit=gen.touched_left_limit,
            touched_right_limit=gen.touched_right_limit, 
            touched_width_limit=gen.touched_width_limit, 

            left_low_depth=gen.left_low_depth,
            right_low_depth=gen.right_low_depth,
            left_low_MQ=gen.left_low_MQ,
            right_low_MQ=gen.right_low_MQ,

            left_okay=left_okay, 
            right_okay=right_okay,
            edited=edited,
        )

    def prepare_vcfspecs(self):
        self.set_row_specs()
        self.set_row_spec_groups()
        self.save_superseq_alignments(raise_with_tie=False)
        self.set_contig_vcfspecs()

    def secure_vcfspec_margins(self):
        vcfspec_margins_gr = self.get_vcfspec_margins_gr(
            subseq_portion_threshold=self.params['allele_portion_threshold'],
            inverse=False,
            split_contig_vcfspec=True,
        )

        if vcfspec_margins_gr.empty:
            # When there is no contig vcfspec
            return realign_base.SecureResult(
                touched_left_limit=self.check_touches_left_limit(),
                touched_right_limit=self.check_touches_right_limit(), 
                touched_width_limit=self.check_touches_width_limit(), 

                left_low_depth=self.check_left_depth_low(),
                right_low_depth=self.check_right_depth_low(),
                left_low_MQ=self.check_left_MQ_low(),
                right_low_MQ=self.check_right_MQ_low(),

                left_okay=True, 
                right_okay=True,
                edited=False,
            )

        # set parameters
        desired_start0 = min(vcfspec_margins_gr.Start)  #min(candidate_start0s)
        #desired_end0 = max(candidate_end0s)
        desired_end0 = max(vcfspec_margins_gr.End)
        left_okay = self.start0 <= desired_start0
        right_okay = self.end0 >= desired_end0

        # When requirements are already met
        if left_okay and right_okay:
            return realign_base.SecureResult(
                touched_left_limit=self.check_touches_left_limit(),
                touched_right_limit=self.check_touches_right_limit(), 
                touched_width_limit=self.check_touches_width_limit(), 

                left_low_depth=self.check_left_depth_low(),
                right_low_depth=self.check_right_depth_low(),
                left_low_MQ=self.check_left_MQ_low(),
                right_low_MQ=self.check_right_MQ_low(),

                left_okay=left_okay, 
                right_okay=right_okay,
                edited=False,
            )

        # begin looping
        initial_start0 = self.start0
        initial_end0 = self.end0
        gen = self.get_extend_generator()
        while True:
            if self._securehelper_check_aborted(gen, left_okay, right_okay):
                break

            if not self._securehelper_check_left_blocked(gen, left_okay):
                gen.iter_left()
                left_okay = gen.current_start0 <= desired_start0

            if not self._securehelper_check_right_blocked(gen, right_okay):
                gen.iter_right()
                right_okay = gen.current_end0 >= desired_end0

        if (initial_start0 != gen.current_start0) or (initial_end0 != gen.current_end0):
            edited = True
            self.subset(gen.current_start0, gen.current_end0, inplace=True)
        else:
            edited = False

        return realign_base.SecureResult(
            touched_left_limit=gen.touched_left_limit,
            touched_right_limit=gen.touched_right_limit, 
            touched_width_limit=gen.touched_width_limit, 

            left_low_depth=gen.left_low_depth,
            right_low_depth=gen.right_low_depth,
            left_low_MQ=gen.left_low_MQ,
            right_low_MQ=gen.right_low_MQ,

            left_okay=left_okay, 
            right_okay=right_okay,
            edited=edited,
        )

    def get_margin_from_vcfspec(self, vcfspec):
        leftmost = vcfspec.leftmost()
        rightmost = vcfspec.rightmost()
        padding = int(len(leftmost.REF_range0) * self.params['vcfspec_range_factor'])
        return (leftmost.start0 - padding, rightmost.end0 + padding)

    # pyranges
    def get_gr(self):
        return pr.PyRanges(chromosomes=[self.chrom], starts=[self.start0], ends=[self.end0])

    def get_active_info_gr(self):
        chroms = list()
        start0s = list()
        end0s = list()
        is_active = list()
        for key, subiter in itertools.groupby(
            self.active_info.items(), key=(lambda x: x[1]),
        ):
            subiter = tuple(subiter)
            chroms.append(self.chrom)
            start0s.append(subiter[0][0])
            end0s.append(subiter[-1][0] + 1)
            is_active.append(key)
        return pr.from_dict(
            {'Chromosome': chroms, 'Start': start0s, 'End': end0s, 'Active': is_active}
        )

    def get_vcfspec_margins_gr(self, subseq_portion_threshold, inverse=False, split_contig_vcfspec=True):
        """Args:
            split_contig_vcfspec: If False, spaces between component vcfspecs of a contig vcfspec are incoporated into result pyranges.
            inverse: If True, subtraction from whole pileup region to vcfspec region is returned.
        """
        start0s = list()
        end0s = list()
        names = list()
        # collect vcfspec margins
        for row_id, contig_vcfspec in self.iter_contig_vcfspecs(subseq_portion_threshold):
            if contig_vcfspec is None:
                continue

            # make vcfspec iterator 
                # when contig vcfspec is composed of only one component vcfspec, 
                # iterator must be set differently
            if len(contig_vcfspec.components[1]) == 0:
                iterator = [contig_vcfspec]
            else:
                iterator = contig_vcfspec.components[1]

            if split_contig_vcfspec:
                for vcfspec in iterator:
                    margins = self.get_margin_from_vcfspec(vcfspec)
                    start0s.append(margins[0])
                    end0s.append(margins[1])
                    names.append(vcfspec.get_id())
            else:
                # for each contig vcfspec, spaces between component vcfspecs are
                # included in the result
                start0_candidates = list()
                end0_candidates = list()
                for vcfspec in iterator:
                    margins = self.get_margin_from_vcfspec(vcfspec)
                    start0_candidates.append(margins[0])
                    end0_candidates.append(margins[1])
                start0s.append(min(start0_candidates))
                end0s.append(max(end0_candidates))
                names.append(contig_vcfspec.get_id())

        # make pyranges object
        if len(start0s) == 0:
            vcfspec_margins_gr = pr.PyRanges()
        else:
            chroms = [self.chrom] * len(start0s)
            vcfspec_margins_gr = pr.from_dict(
                {'Chromosome': chroms, 'Start': start0s, 'End': end0s, 'Name': names}
            )
        # return; inverse if needed
        if inverse:
            return self.get_gr().subtract(vcfspec_margins_gr)
        else:
            return vcfspec_margins_gr


    # secure_inactive_padding
#    def secure_inactive_padding_rightward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
#        secure_inactive_padding_rightward(
#            pileup=self,
#            inactive_padding=inactive_padding, 
#            extend_pileup_by=extend_pileup_by, 
#            inplace=True,
#        )
#
#    def secure_inactive_padding_leftward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
#        secure_inactive_padding_leftward(
#            pileup=self, 
#            inactive_padding=inactive_padding, 
#            extend_pileup_by=extend_pileup_by, 
#            inplace=True,
#        )
#
#    def secure_inactive_padding(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
#        self.secure_inactive_padding_rightward(inactive_padding, extend_pileup_by=extend_pileup_by)
#        self.secure_inactive_padding_leftward(inactive_padding, extend_pileup_by=extend_pileup_by)

    ### row specs ###
    def set_row_specs(self):
        self.row_specs = dict()
        for read_uid, row in self.dfs['queryonly_added_main'].iterrows():
            #row = self.df.loc[read_uid, :]
            self.row_specs[read_uid] = self._make_row_spec(row)

    def row_spec_getter(self, rowid):
        return self.row_specs[rowid]

    def set_row_spec_groups(self):
        """row_spec's are grouped by whether one is a substring of another."""
        self.row_spec_groups = collections.OrderedDict()

        for query_rowid, query in sorted(
            self.row_specs.items(), 
            #key=(lambda x: (len(x[1]['seq']), x[1]['span_length'])), 
            #key=(lambda x: (x[1]['span_length'], -len(x[1]['seq']))), 
            key=(lambda x: (x[1]['span_length'], -x[1]['cliplen'])), 
            reverse=True,
        ):
            superseq_candidates = list()
            for superseq_rowid in self.row_spec_groups.keys():
                target = self.row_specs[superseq_rowid]
                if self.row_spec_matcher(query, target):
                    superseq_candidates.append(superseq_rowid)
                #if len(superseq_candidates) >= 2:
                #    break
                    
            if len(superseq_candidates) == 0:
                self.row_spec_groups[query_rowid] = {
                    'subseq_rowids': list(),
                    'subseq_hits': 0,
                }
                self.row_spec_groups[query_rowid]['subseq_rowids'].append(query_rowid)
                self.row_spec_groups[query_rowid]['subseq_hits'] += 1
            else:
                superseq_rowid = self.superseq_candidates_tiebreak(superseq_candidates, self.row_spec_getter, self.row_spec_groups)
                groupinfo = self.row_spec_groups[superseq_rowid]
                groupinfo['subseq_rowids'].append(query_rowid)
                groupinfo['subseq_hits'] += 1

        self.decorate_groupinfo(self.row_spec_groups, self.read_getter)

    def read_getter(self, key):
        return self.read_store[key]

    # alignment #
    def align_row_spec_to_ref(self, row_spec, ref_seq, ref_seq_reversed, raise_with_tie):
        if row_spec['seq'] == '':
            return Bio.Align.Alignment(
                sequences=[ref_seq, ''],
                coordinates=np.array([(0, len(ref_seq)), (0, 0)]),
            )

        reverse_align = (not row_spec['left_filled']) and row_spec['right_filled']
        target = ref_seq
        query = row_spec['seq']
        return realign_base.main_aligner(
            target, 
            query, 
            self.aligner, 
            reverse_align=reverse_align, 
            raise_with_tie=raise_with_tie, 
            logger=self.logger, 
            row_spec=row_spec, 
            target_reversed=ref_seq_reversed,
        )

    def align_row_spec_to_ref_old(self, row_spec, ref_seq, ref_seq_reversed, aligner, raise_with_tie):
        # set params
        query = row_spec['seq']
        reverse_align = (not row_spec['left_filled']) and row_spec['right_filled']
        # run aligner
        try:
            if reverse_align:
                alns = aligner.align(ref_seq_reversed, query[::-1])
            else:
                alns = aligner.align(ref_seq, query)
        except Exception as exc:
            raise Exception(f'Failed alignment:\nrow_spec: {row_spec}\nReference seq: {ref_seq}') from exc

        self.logger.debug(f'Num of alignments: {len(alns)}; row_spec: {row_spec}')

        # treat dirty alignments
        if len(alns) > 10000:
            aln = random.choice(alns)
            if reverse_align:
                aln = alignhandler.reverse_alignment(aln)
            #aln = alignhandler.amend_outer_insdel_both(aln)
            self.logger.debug(f'Skipped dirty alignment; row_spec: {row_spec}')
        else:
            # recover reversed alignment
            if reverse_align:
                alns = [alignhandler.reverse_alignment(x) for x in alns]

            if len(alns) == 1:
                aln = alns[0]
                aln = alignhandler.amend_outer_insdel_both(aln)
            else:
                #alns = [alignhandler.amend_outer_insdel_both(x) for x in alns]
                try:
                    aln = self.align_row_spec_to_ref_helper(alns, raise_with_tie, row_spec)
                except TimeoutError:
                    self.logger.debug(f'skipping alignments tiebreaking due to timeout;\nrow_spec: {row_spec}')
                    aln = alns[0]
                    aln = alignhandler.amend_outer_insdel_both(aln)
            
        return aln

    @deco.get_deco_timeout(0.05)
    def align_row_spec_to_ref_helper(self, alns, raise_with_tie, row_spec):
        alns = list(
            alignhandler.remove_identical_alignments(
                alignhandler.amend_outer_insdel_both(x) for x in alns
            )
        )
        try:
            aln = alignhandler.alignment_tiebreaker(alns, raise_with_failure=raise_with_tie)
        except alignhandler.AlignmentTieError as exc:
            msg = f'Failed to break alignment tie. row_spec is:\n{row_spec}'
            raise Exception(msg) from exc

        return aln

    def save_superseq_alignments(self, raise_with_tie=False):
        """'_superseq_alignments' attribute is set."""
        self._superseq_alignments = dict()
        ref_seq = self.get_ref_seq()
        ref_seq_reversed = ref_seq[::-1]
        for superseq_rowid in self.row_spec_groups.keys():
            row_spec = self.row_specs[superseq_rowid]
            self._superseq_alignments[superseq_rowid] = self.align_row_spec_to_ref(row_spec, ref_seq, ref_seq_reversed, raise_with_tie)

    # not used #
    def save_subseq_alignments(self):
        self._subseq_alignments = dict()
        for superseq_rowid, groupinfo in self.row_spec_groups.items():
            sup_to_ref_aln = self._superseq_alignments[superseq_rowid]
            for subseq_rowid in groupinfo['subseq_rowids']:
                subseq_row_spec = self.row_specs[subseq_rowid]
                self._subseq_alignments[subseq_rowid] = self.align_subseq_to_ref(subseq_row_spec, sup_to_ref_aln)

    # vcfspec #
    def set_contig_vcfspecs(self):
        """- Must be run after row_spec_groups is set and save_superseq_alignments have been run
        """
        self.contig_vcfspecs = dict()
        for superseq_rowid, groupinfo in sorted(
            self.row_spec_groups.items(),
            key=(lambda x: x[1]['subseq_hits']),
            reverse=True,
        ):
            superseq_row_spec = self.row_specs[superseq_rowid]
            superseq_alignment = self._superseq_alignments[superseq_rowid]
            superseq_vcfspec = self.superseq_to_vcfspecs(superseq_row_spec, superseq_alignment)
            self.contig_vcfspecs[superseq_rowid] = superseq_vcfspec

#    def get_realigned_reads(self):
#        """Must be run after 'save_superseq_alignments' method
#        Returns:
#            dict (keys ReadUID, values pysam.AlignedSegment)
#        """
#        realigned_reads = dict()
#    #    for superseq_rowid, groupinfo in sorted(
#    #        active_region_pileup.row_spec_groups.items(),
#    #        key=(lambda x: x[1]['subseq_hits']), 
#    #        reverse=True,
#    #    ):
#        # sorting is not necessary
#
#        for superseq_rowid, groupinfo in self.row_spec_groups.items():
#            #realigned_read_superseq = realign_superseq(superseq_rowid, active_region_pileup, active_range)
#            #realigned_reads[superseq_rowid] = realigned_read_superseq
#                # (221019) This is commented out because superseq itself is
#                # included in subseq list.
#            superseq_aln = self._superseq_alignments[superseq_rowid]
#            for subseq_rowid in groupinfo['subseq_rowids']:
#                #if subseq_rowid in realigned_reads.keys():
#                #    continue
#                    # (221019) This is commented out because subseqs with multiple superseq hits are
#                    # incoporated into only one superseq group.
#                realigned_read_subseq = realign_subseq(subseq_rowid, self, self.range0, superseq_aln)
#                realigned_reads[subseq_rowid] = realigned_read_subseq
#                
#        return realigned_reads

    # split point generation
    def get_split_points(self):
        """Result may be empty"""
        # without splitting contig vcfspec
        splittable_region = self.get_vcfspec_margins_gr(
            subseq_portion_threshold=(self.params['allele_portion_threshold'] * 2),
            inverse=True,
            split_contig_vcfspec=False,
        )

        if splittable_region.empty:
            return list()
#            # split contig vcfspec
#            splittable_region = self.get_vcfspec_margins_gr(
#                subseq_portion_threshold=(self.allele_portion_threshold * 2),
#                inverse=True,
#                split_contig_vcfspec=True,
#            )
#            if splittable_region.empty:
#                logging.info(f'Getting Pileup splitting point from inactive runs because splitting with vcfspec margins failed. Pileup object being split: {self}')
#                # from inactive runs
#                #split_points = [self.get_split_point_from_inactive_runs()]
#                return list()
#            else:
#                logging.info(f'Getting Pileup splitting point, splitting contig vcfspecs. Pileup object being split: {self}')
#                pass
        else:
            pass

        split_points = list()
        for idx, row in splittable_region.df.iterrows():
            point = int((row.Start + row.End) / 2)
            if point != self.start0:
                split_points.append(point)

        return split_points

    def get_split_point_from_inactive_runs(self):
        """Midpoint of the widest non-marginal inactive area"""
        active_info_gr = self.get_active_info_gr()
        inactive_gr = active_info_gr[~active_info_gr.Active]

        if inactive_gr.empty:
            raise realign_base.SparseInactiveRegionError(f'{self}')

        max_width_index = inactive_gr.lengths().argmax()
        max_width_row = inactive_gr.df.iloc[max_width_index, :]
        return int((max_width_row.Start + max_width_row.End) / 2)

    def get_result_vcfspecs(self, as_components=False, merge=True, subseq_portion_threshold=None, MQ_threshold=40): 
        if subseq_portion_threshold is None:
            subseq_portion_threshold = self.params['allele_portion_threshold']

        if as_components:
            result = list()
            for row_id, contig_vcfspec in self.iter_contig_vcfspecs(subseq_portion_threshold=subseq_portion_threshold, MQ_threshold=MQ_threshold):
                if contig_vcfspec is None:
                    continue

                if len(contig_vcfspec.components[1]) == 0:
                    result.append(contig_vcfspec)
                else:
                    result.extend(contig_vcfspec.components[1])
        else:
            result = list(
                x[1] for x in self.iter_contig_vcfspecs(subseq_portion_threshold, MQ_threshold)
                if x[1] is not None
            )

        if merge:
            if len(result) > 0:
                result = functools.reduce(lambda x, y: libvcfspec.merge(x, y), result)

        return result

    # for debugging
    def show_row_spec_alignments(self, skip_zero_hits=True, show_subseqs=False):
        self._show_row_spec_alignments_helper(
            self.row_spec_groups, 
            lambda x: self.row_specs.__getitem__(x),
            lambda x: self._superseq_alignments.__getitem__(x),
            skip_zero_hits, 
            show_subseqs,
        )





import textwrap

import pysam

import handygenome.workflow as workflow
import handygenome.vcfeditor.indexing as indexing


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser(
        description=f'Makes a new vcf file where all INFO and FORMAT annotations are removed. Relevant header metadata lines are also removed.'
    )

    workflow.add_infile_arg(parser_dict['required'], required=True)
    workflow.add_outfile_arg(parser_dict['required'], required=True)
    workflow.add_outfmt_arg(parser_dict['optional'], required=False, default='z')
    workflow.add_index_arg(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def make_new_header(in_vcf):
    new_header = pysam.VariantHeader()
    for chrom, contigitem in in_vcf.header.contigs.items():
        new_header.contigs.add(contigitem.name, contigitem.length)

    return new_header


def write_new_vrs(in_vcf, out_vcf):
    for vr in in_vcf.fetch():
        new_vr = out_vcf.new_record()
        for key in ('contig', 'pos', 'id', 'ref', 'alts'):
            setattr(new_vr, key, getattr(vr, key))
        out_vcf.write(new_vr)


def main(cmdargs):
    args = argument_parser(cmdargs)
    in_vcf = pysam.VariantFile(args.infile_path)
    new_header = make_new_header(in_vcf)
    out_vcf = pysam.VariantFile(args.outfile_path, mode=args.mode_pysam, header=new_header)
    write_new_vrs(in_vcf, out_vcf)
    out_vcf.close()
    in_vcf.close()
        
    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)



import handygenome.workflow as workflow
#import handygenome.workflow.toolsetup as toolsetup
import handygenome.dbconv.cosmicdb as cosmicconv


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(
            f'COSMIC database files "CosmicNCV.tsv.gz" and '
            f'"CosmicMutantExport.tsv.gz" are used to create a single vcf '
            f'file which can be used for annotation.'))

    workflow.add_outfile_arg(
        parser_dict['required'], required=True, must_not_exist='ask',
        help='Output vcf file path.')

    parser_dict['required'].add_argument(
        '--CosmicNCV', dest='ncv_path', 
        required=True, 
        help=(f'Previously downloaded unmodified CosmicNCV.tsv.gz file'))

    parser_dict['required'].add_argument(
        '--CosmicMutantExport', 
        dest='mutexp_path', 
        required=True, 
        help=(f'Previously downloaded unmodified CosmicMutantExport.tsv.gz file'))

    parser_dict['required'].add_argument(
        '--refver', dest="refver",
        required=True, choices=('GRCh37', 'GRCh38'),
        help=f'Reference genome version. Must be one of "GRCh37" or "GRCh38".')
        
    parser_dict['required'].add_argument(
        '--cosmic-version', dest="cosmicver",
        required=True,
        help=f'COSMIC database version (e.g. v96).')

    workflow.add_outfmt_arg(parser_dict['optional'], required=False)
    workflow.add_logging_args(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)
    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    #logger = toolsetup.setup_logger(args)
    cosmicconv.main(args.ncv_path, args.mutexp_path, args.outfile_path, args.refver, args.cosmicver)




#!/usr/bin/env python3

#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
import os
import sys
import re
import subprocess
import argparse
import time
from scipy import spatial
import pandas as pd
import shutil
import logging
import hashlib
from handygenome.sigprofiler_clone.SigProfilerMatrixGenerator.scripts import SigProfilerMatrixGeneratorFunc as matGen
import glob

check_sum = {'GRCh37':
						{'1':'a7d51305e943cf06ff2029146bd91bca','2':'d24d0185af89356d44614ab0d6fd6a68','3':'ea5e033147dcaf77bfd4c70f50688d37',
						 '4':'00d7797c7184f1802367e33f6e2bc3da','5':'f74b1eeb329088242a9f22a16322b325','6':'b353cc4c4abc90340e7747509fe7b457',
						 '7':'bbadde91b3ef958c7d20e2b1088c8cd2','8':'0ff695692e3efebaf00c7905d0d536d7','9':'40b75a18acb66748a37888c53a76dcdb',
						 '10':'557881b744a932b4ceee8a06d6de85a4','11':'f8b8f118101d7cb04164b29a7acadca4','12':'52c18e9fefc3ed3e35c1d8771d1247de',
						 '13':'a241d1cdcadccfd94db792300ab000bf','14':'ed3907128336795669bc19d77c0aa409','15':'bfc66ad087c4e9025076d7571cffa30e',
						 '16':'bd251fddc42400bb54ef95d5e1002ece','17':'fcd36b1bf5c4bd74328dc4caaae244ae','18':'e015d4324c36374827582c5b1214a736',
						 '19':'5cfa7d47e2d73dbdbf8d68f97c8e8b23','20':'2fa0717bf4e8dddac64cd393f4134ff5','21':'ba5559776d4601b80ca42c82f02102a4',
						 '22':'ba762b6ae493df40d04d1ff63d9b2933','Y':'0303100be91874b966a998273cd7c8eb','X':'14e331d82736f6cfc177ff8c90f7bd78',
						 'MT':'dfd6db5743d399516d5c8dadee5bee78'},
			'GRCh37_havana':
						{'1':'a954558ed4a92c33454f8539d5e47d43','2':'f2a1702c668d531656ffbeadb5bb1b66','3':'e6a4d10c433d8707a1a336e310a779bc',
						 '4':'709d6d8ab80b8f83f2f8892e23a9adba','5':'fa4fe82dce241cf4eeaf84ef9b44030f','6':'161d68063347fff0fe5d82a1c60cdef4',
						 '7':'5f866a10820d3d48d3b96fb8432ade8e','8':'585c7100d1de569df883910bbf01463c','9':'5667fc1a44a93b6f162bf60d4f6a21c1',
						 '10':'7684e3198ac4c484b56d6e91b8f2fc15','11':'e3b5501a11be7d5fcf9219933467c86a','12':'511a8b89d19e47df4633da5f277373b5',
						 '13':'7b621f0c18db0592c0f82db9b760ceef','14':'16b383fa1dbf30feb8a5d1d89a9ac7af','15':'f3295833a4dfd7a628ca9befe1972739',
						 '16':'9985d8cf725dcd29b3ded0a4eeb426e2','17':'9118e1200b7cd95a4d5c95ff0e039c7c','18':'116ae0ba70ea43b10983ddd07387d6dd',
						 '19':'930c568fe249b78a602a3400ce094b83','20':'fa24f27c160dffd5d8b7c6270dff4a0d','21':'4f6f61d43c698e2c66dc62f21f7ad6cd',
						 '22':'77661703e7b65ad2b6adcb75e7e3e53b','Y':'b86042fd443490fb0061478037392fc0','X':'02b7328d7d74704d571fd38149bbf814',
						 },

			'GRCh38':
						{'1':'ebe083105e7703a49581a36d73732a96','2':'cd65e36dbdf12a8ac3d2c70ebac8cad4','3':'6c20a7008394f2fa9c304d231a1f391b',
						 '4':'5c7443e1678868adadeac0e57558f6e8','5':'45573232c8097c679503a6598f61e60b','6':'cfc137c7434d3a9a872332d405b5c553',
						 '7':'9d8210c22c1962db837e7b62a578975c','8':'665134fd44f21915cbeef955addf89ba','9':'758d0c0c71d8bafbe1ede86587191730',
						 '10':'397bb21acff1ca3052ac802f2aee06e0','11':'07707ff8a2a964656469a7be7bb3e576','12':'506d02539075e080ee12ebdf63908080',
						 '13':'03ed22f01ab43145733c0b6a647e0560','14':'8b93447086549e476c65699ed813a567','15':'cd0dfe9fa78cae2fc7becf8f8ec6c693',
						 '16':'e17bbb66eb4d6b62b7b0e2fbf062b6a6','17':'8fc95bb3101d024d890aa3543eb454c5','18':'a4870628045bb033a90e8c89f818e24d',
						 '19':'6a9d0c8298f0ba2fa13180e02b969f16','20':'aa75d35969cf3956bb4ace7bdc57b34e','21':'5d55f5ad6271d6a0d8806876924990f7',
						 '22':'efdb4e1d23ab7964302b828062a33447','Y':'3b38c639ad164d60f1a055b46fcd2748','X':'d5edbea3cf5d1716765dd4a7b41b7656',
						 'MT':'dfd6db5743d399516d5c8dadee5bee78'},

			'GRCh38_havana':
						{'1':'c4ef4ee14a4f0f7b319e9ed01f2a9742','2':'189cb33da673afcf161b7724d78d314b','3':'19fb72b2f4bebca66969d9eaaab0b06b',
						 '4':'04069627d35462add3095cd14299c181','5':'e3fbab13e69a5ba48c9bf706215aef19','6':'7de52b0136f65c25b2f3652e8888449a',
						 '7':'0071cc7e4024b0d81941df353ce99fdc','8':'c19f322ee6a786e94e665b1ab3f459ea','9':'00920f091df48d3dba8b270432bb38a8',
						 '10':'dfcdfcc0a2524888e91f9a40aa2289e5','11':'01118568e33f876f31f4d0585085b844','12':'11b6ffc45b5d198344e8d0f2e51fa08a',
						 '13':'e1241284e80f9a3afe204131c9eb6192','14':'b8d68bd919dc87483679314d0dae9514','15':'ca2c6d19dadc4d905013b37dfaa75fa6',
						 '16':'ccf7fe495c7ff7c11c430a935050bb4d','17':'ec09c6c67cc0496117da52bc584cab8f','18':'e8edcd9275c659101745570b0ddc436d',
						 '19':'293befa3316b6bb8b5a2a2c75e5e8030','20':'4065b4403e1b9659f9f049c4c886ed7f','21':'3f661bcfd29bf7153b4bfb95e8d112d8',
						 '22':'ea31a6d2cdac7d8aeb2e3be0688b4159','Y':'a68939c10c3ebc16340785bf99366328','X':'57623734b88f441f5499955d2c83a6f9',
						 },

			'mm9':
						{'1':'c5afc4b3f7f2119696214511d7a04341','2':'a7b467475a1b032d2c893dac1c419a28','3':'f922bc529a17324f1cd858f9a8723d65',
						'4':'f3d6b74e3c04dbd229e2f1e363607506','5':'5fee4f1889c9fe20f7f8562c62bbeb0a','6':'481d47b87da45f3a20181c780fd796c2',
						'7':'454ef2bf49a5ba8cfea3d16dfcfc7f25','8':'2f4162d4c824db78a2a2a820cb4fec81','9':'0649e6aec61af1ab8ab4797ea8e54119',
						'10':'38296256bcfe886c8ae771418e4fd824','11':'b31cb0ce693e35eaa77031d44b12e474','12':'d2b3e4b015742b6aea30ceec5a972968',
						'13':'df77b6d0ed1b133224b128c189736372','14':'0ec3c0e6b3fa2cdb957541f19792e130','15':'44fcaf2ec9b82dae910f85ce41c3cfad',
						'16':'ad7a8dbdf46fa7077e0982a54eab70b7','17':'71aee1dee3cd2078e4619c485d88817e','18':'727ec4ed3128ecacd6cd2f7558083553',
						'19':'461a7119781ab7f4b654fdd9ef76e0ec','Y':'471ff3bbb4520c020cfaa7ca8371c543','X':'9ccadf96cd3aa0ed9d299894a3d7fde0',
						'MT':'a1d56043ed8308908965dd080a4d0c8d'},

			'mm10':
						{'1':'ef88c5ac276a32a2865c0408f92acd55','2':'ced7325ef9e2dfedea3fbe26428a6059','3':'9cd1794eeea27553077a018038303908',
						'4':'da616d7ed6c67f824487eb2ed09cd33b','5':'b327b82da6986bf947105d07c0ad6d2e','6':'fb9a8fa0b85561f8d4de633c22d5157a',
						'7':'12457fd80f6806779fc0d4cc8d36fbad','8':'5d98d86bd22bee1cb226406f49ee7caf','9':'b2f26613fcc622a4003e4c945ae55e25',
						'10':'e9f3589529e258ede66d2e77bb87d21d','11':'76bcd285c3c66471ad6fccfabe42294c','12':'ac34fc3616c9609d8e75a59069e9007a',
						'13':'f81b976e4e4617b25945d06f9aa30846','14':'95dc042eb2aa7d4cc0abe071d4d7966e','15':'fbf2477833aff73ae085537cd7ee0f85',
						'16':'77cbcd009ba50891571f785595717ec1','17':'cd9e4dfdd168ed3de05dac4d44c6e692', '18':'945e83694c7c8f69d6186e1a2abc9771',
						'19':'e57b25f8869de31a9dbce06510711db6','Y':'c2146ba4ab1ec262f5e38b2a1ebc5f5b','X':'9af543088be046fdc63976c2d41de94c',
						'MT':'a1d56043ed8308908965dd080a4d0c8d'},
			'mm39':
						{'1':'57c0e2508c3f8fe66a792be344f494d9','2':'8e69643ce8f9ff048eae0113f64ce8b7','3':'0a1f33496316ecc245868ca72dc4ac95',
						'4':'15f59279c79b0b23cda1cac00aec9b2f','5':'cd6709c54421184266eea49be266378c','6':'916484773b14295b4711e9f13efd1c94',
						'7':'04a6a77f0754a0b6badd825c0e8729ff','8':'2f68f81e848940f3e8e5e4e2e9f5c6bc','9':'f2cefa27e8d26e588d32212c78c5b840',
						'10':'5f877fd57657884cdc3c09036a365314','11':'1b7b5162cc876b2fb0502e010daf1a50','12':'cfccf49fc2034eebd66ea40e2410c482',
						'13':'a411479e72e6ee2fc9b1667a7598de0d','14':'11dfe2f88d544c1402be18bc732a7b80','15':'9d71bd08f73fe51d24856c804cb86b38',
						'16':'9b97d374745941c370e35596846e2738','17':'4a698fcd2e16eea7d7a718c94747e85d', '18':'fe0a2d35beb0407114a9e61b6ed646f2',
						'19':'c51fa76818b0447e13047095251796eb','Y':'554d0a9f1062fb12903e521c866e2600','X':'ce0f276bc6eab955c22a0111bc436ad7'},
			'mm10_havana':
						{'1':'8351fa274b2e49b159fdfd1a17b8e522','2':'afa7459e3fa1a40f96279d318e3f6fc4','3':'106064694826ed8b9cfd29c5cfe39f1d',
						'4':'807aafbf65a28995b2ff64c040f54ff8','5':'afc3c4640788db5b8866d7644be4c49d','6':'2eff5f397f61735e90dc4089705751bc',
						'7':'f62fe7d720c219bacbe91fb0624fdb33','8':'1e523de22029fc4feea1e3222638d3ec','9':'3b75a1f2360ef49cbdc3da16a1e40d23',
						'10':'2d3cbd3aa1b5f6c78a11bc799787d226','11':'9b5e62ac65c2e05e7c09d7b3ac9aa507','12':'07e2bfe001933efb9052458f203fd8d4',
						'13':'4df084dfa4d2ea80519da4fa825d7c6e','14':'c4c01a321c2e8b67ed619dc91588d1cc','15':'6ccda29f891ac9d17567c69a1b47a40a',
						'16':'7df7554d6b3d7ca8331fe163e4930a64','17':'60ecd93b145f36226c246356701422f2', '18':'b52767b6bbc3a293993652e78db7ddac',
						'19':'5a246e5963dc0a300765b5cdd87e07d9','Y':'da73d749cd65735db5fd10556ae2376a','X':'ab9c1ad518e71b15ff543cda415a463f',
						},

			'rn6':
						{'1':'003723513cbdb3708fcc5d737c05199c','2':'53e52c5facc7f05462be533845f37425','3':'8d157a9b71fe9770cf783ea5459b19d7',
						'4':'a66dc1999bcc960ff11fe0b24c0d7b14','5':'601cf83411234adbdd9f911b89509564','6':'03b1f4af58fffdf213466ea85b570b3d',
						'7':'4ed05ddf9502ef79e121c02e391660e6','8':'3e2458daaf1b3e8ab4d0e0a9e60c067b','9':'8f83caeccec7ea6e35e404737138ee67',
						'10':'9c1af453a5facc9bfa821457bcfc4d30','11':'ef0480a905c55d76a3c58e295a85bc75','12':'643b6fe4a3a6363ffe64a6c316fa3e1a',
						'13':'102bb3fb420a4104c216bcdf99870374','14':'e26b8b63fba0ea7ced4f0330e93a8cdc','15':'da747616a1362d374d4786102fab6f9f',
						'16':'54e4f932eb0eda4cbf31156f96ef7235','17':'46c2facf5415e4eff8b0804161db722d', '18':'f1cb84f002967854b83bf266ec59a7a3',
						'19':'b85ca155fd1780fe5c327a4589c212a6','20':'899d3511352d78b9b9dc63f063d91b31','Y':'6a7a3539c329dc540dfa6db006003bb1',
						'X':'7a06bafab97c59a819f03633f0a6b7a2'},

			'c_elegans':
						{'I':'5a3ea8cf3dfbc641716b7bc805edcaae','II':'bf82edaa92809dd2fea2b791c38c9728','III':'d2df34b6743f41d3964549fc76c5f1a2',
						'IV':'23396bb57145d3acde2888947b5b8c3a','V':'09df3c53b12e5fd7d9035cc98ca221a3','X':'988046456f1409dfdb5e26444d84d238',
						'MtDNA':'48983f530959780de0125f74a87d4fc1'},

			'dog':
						{'1':'bef8283c1a36f9aef0e407de2ff6af00','2':'9cc961192bb5e58b3847060c3e9c1cfc','3':'d33263fa2de6666b41e140cb7a8da66c',
						 '4':'cd4ed39ebac1c04800ccf30466ec69f5','5':'c0f48a4a764e58388b48835aca2ec0a4','6':'4b472a2f8d0a53ac75cce04e7dc9279a',
						 '7':'12a61573a0da2c9306fff705bb1c39c1','8':'e22cf22a27560aa8523dc959ddcf6e25','9':'c079a73d719145cdd5c7c93969a1c392',
						 '10':'45805a518147f7846bd0457ca038c8df','11':'f38cda8508463a7607dff14a581ee7b0','12':'adb5de197f58bb827fa01fe924eb3a1d',
						 '13':'055a845ba97baad3b13d4d3359f88290','14':'27f0ba8e47996a058807a3827cf8e4a8','15':'2e9565c687a593eb0acbdd0962bb9255',
						 '16':'89b2225bb78d88b0fd1d38d9514ab0cb','17':'f0378253e2f083e42b665ea202fde3b0','18':'04d124e273f3b54a685ad6526223cd03',
						 '19':'67bae093919e6bb5ab6b9806c739d539','20':'5588387165a2e19c4533012cfb4998f3','21':'371cdf18a545728f7964b9db2fc72d5e',
						 '22':'fbf76865f88a018d93506e036f6a68bc','23':'085145e01d9fd9f0f999fb9e8e8d4400','24':'69b75a9962fb766b447e7d1252cb31ac',
						 '25':'12d5c6677b3e17170c317c1f5532d2a8','26':'13937d18e56b2b93d12fa5fcba48a138','27':'1d03d8ca5f201f4d156f5e1b38f7a67c',
						 '28':'c33395dec7fdc13e9d8f10afaa946f8c','29':'174f2db104ecaa5efef770f44241e3b0','30':'047d420ef9aecb933a7d83b6af820b23',
						 '31':'5be61f0c9944a5f2d7d1a5b2e75fb000','32':'212dcb867e95a642277a243fed8d8e41','33':'08a217b02cdd778cfdb0005dff4828b1',
						 '34':'4245d6fc370d9049ef4c25314fbef239','35':'1344aba8755b8a4e304629180fc0591a','36':'e4fff6ed84777905dc999ca6d6bc2557',
						 '37':'60d51ea6ae9e3f2fa316e3d03aff96b2','38':'4090ff76d94e6b38920916ae3ff2441c','X':'bce1372df64037d79b0995311d8ff971'},
			'ebv':
						{'gi_82503188_ref_NC_007605':'7f8894e52b0cac1f968c0402838bea07'},

			'yeast':
					{'I':'c7cd4d6148ff54d1d003d7fe3c39bfdf', 'II':'3303648ecf106df7c6c1449cb55980e8', 'III':'12f784db9c1cd2bc632fbefdb5329504',
					 'IV':'e5bbd701bae1ab3965da90e1131f5b5c', 'V':'b4447ed088b10a404925c153ca3055d4', 'VI':'3285962c283977064340b62a7f843601', 
					 'VII':'72e1438a61fde52bd66cef12c53ae4db', 'VIII':'0e00648307fd4d64f50d0feb7e428e13', 'IX':'43466267caa5ae7aec23df022b583564',
					 'X':'dac706139aeb5cd3d75f24d0e405131e', 'XI':'5d4b415e1e313918bba6cd061d416d58', 'XII':'e26e59a4b4944ab96b819a92b60c9868',
					 'XIII':'a12878f65ed7464e2824cb0c291d359e', 'XIV':'807431901ba993e6ba87861401da1cb2', 'XV':'44fd7223d9d30bdcf94cf627e5a172ca',
					 'XVI':'670d7aad863043ea7a125f004e281483'}}

def get_reference_md5():
	return check_sum

def md5(fname):
	hash_md5 = hashlib.md5()
	with open(fname, "rb") as f:
		for chunk in iter(lambda: f.read(4096), b""):
			hash_md5.update(chunk)
	return (hash_md5.hexdigest())


def install_chromosomes (genomes, ref_dir, custom, rsync, bash):
	if custom:
		for genome in genomes:
			os.system("gzip -d references/chromosomes/fasta/" + genome + "/*.gz")
			chromosome_fasta_path = "references/chromosomes/fasta/" + genome + "/"
			os.system("python scripts/save_chrom_strings.py -g " + genome + " -c " + str(custom))
			print("Chromosome string files for " + genome + " have been created. Continuing with installation.")
			#os.system("rm -r " + chromosome_fasta_path)
	else:
		for genome in genomes:
			species = None
			chrom_number = None
			if genome == 'GRCh37' or genome == 'GRCh38': 
				species = "homo_sapiens"
				chrom_number = 24
			elif genome == 'mm10' or genome == 'mm9':
				species = "mus_musculus"
				chrom_number = 21
			elif genome == 'rn6':
				species = 'rattus_norvegicus'
				chrom_number = 22
			else:
				print(genome + " is not supported. The following genomes are supported:\nGRCh37, GRCh38, mm10")
				sys.exit()
			
			chromosome_string_path = "references/chromosomes/chrom_string/" + genome + "/"
			chromosome_fasta_path = "references/chromosomes/fasta/" + genome + "/"

			if os.path.exists(ref_dir + "chromosomes/tsb/" + genome) and len(os.listdir(ref_dir + "chromosomes/tsb/" + genome)) >= chrom_number:
				break
			wget_flag = True
			if os.path.exists(chromosome_string_path) == False or len(os.listdir(chromosome_string_path)) <= chrom_number:
				print("[DEBUG] Chromosome string files found at: " + ref_dir + chromosome_string_path)
				if os.path.exists(chromosome_fasta_path) == False or len(os.listdir(chromosome_fasta_path)) <= chrom_number:
					print("[DEBUG] Chromosome fasta files found at: " + ref_dir + chromosome_fasta_path)
					print("Chromosomes are not currently saved as individual text files for " + genome + ". Downloading the files now...")
					if not rsync:
					#os.system("rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/grch37/update/fasta/homo_sapiens/dna/ " + chromosome_fasta_path + " 2>&1>> install.log")
						# try:
						#     p = subprocess.Popen("wget", stdout=subprocess.PIPE, stderr=subprocess.PIPE)
						# except:
						#     proceed = input("You may not have wget or homebrew installed. Download those dependencies now?[Y/N]").upper()
						#     if proceed == 'Y':
						#         try:
						#             os.system("brew install wget")
						#         except:
						#             os.system('/usr/bin/ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"')
						#             os.system("brew install wget")
						#     else:
						#         print("Installation has stopped. Please download the chromosome files before proceeding with the installation.")
						#         wget_flag = False
						#         sys.exit()
						if wget_flag:
							try:
								if genome == 'GRCh37':
									if bash:
										os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/grch37/current/fasta/homo_sapiens/dna/ 2>> install.log' + "'")
									else:
										os.system('wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/grch37/current/fasta/homo_sapiens/dna/ 2>> install.log')
									#os.system("wget -r -l1 -c -nc --no-parent -A '*.dna.chromosome.*' -nd -P " + chromosome_fasta_path + " ftp://ftp.ensembl.org/pub/grch37/update/fasta/homo_sapiens/dna/ 2>> install.log")
								elif genome == 'mm9':
									if bash:
										os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-67/fasta/mus_musculus/dna/ 2>> install.log' + "'")
									else:
										os.system('wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-67/fasta/mus_musculus/dna/ 2>> install.log')

								elif genome == 'rn6':
									if bash:
										os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-96/fasta/rattus_norvegicus/dna/ 2>> install.log' + "'")
									else:
										os.system('wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-96/fasta/rattus_norvegicus/dna/ 2>> install.log')
								else:
									if bash:
										os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-93/fasta/' +species+'/dna/ 2>> install.log' + "'")
									else:
										os.system('wget -r -l1 -c -nc --no-parent -A "*.dna.chromosome.*" -nd -P ' + chromosome_fasta_path + ' ftp://ftp.ensembl.org/pub/release-93/fasta/' +species+'/dna/ 2>> install.log')

								#os.system("gunzip references/chromosomes/fasta/" + genome + "/*.gz")
								os.system("gzip -d references/chromosomes/fasta/" + genome + "/*.gz")

							except:
								print("The ensembl ftp site is not currently responding.")
								sys.exit()
					else:
						try:
							if genome == 'GRCh37':
								if bash:
									os.system("bash -c '" + "rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/grch37/current/fasta/homo_sapiens/dna/ " + chromosome_fasta_path + " 2>&1>> install.log" + "'")
								else:
									os.system("rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/grch37/current/fasta/homo_sapiens/dna/ " + chromosome_fasta_path + " 2>&1>> install.log")
							elif genome == 'mm9':
								if bash:
									os.system("bash -c '" + "rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-67/fasta/mus_musculus/dna/ " + chromosome_fasta_path + " 2>&1>> install.log" + "'")
								else:
									os.system("rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-67/fasta/mus_musculus/dna/ " + chromosome_fasta_path + " 2>&1>> install.log")
							elif genome == 'rn6':
								if bash:
									os.system("bash -c '" + "rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-96/fasta/rattus_norvegicus/dna/ " + chromosome_fasta_path + " 2>> install.log" + "'")
								else:
									os.system("rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-96/fasta/rattus_norvegicus/dna/ " + chromosome_fasta_path + " 2>> install.log")
							else:
								if bash:
									os.system("bash -c '" + "rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-93/fasta/"+species+"/dna/ " + chromosome_fasta_path + " 2>&1>> install.log" + "'")
								else:
									os.system("rsync -av -m --include='*/' --include='*.dna.chromosome.*' --exclude='*' rsync://ftp.ensembl.org/ensembl/pub/release-93/fasta/"+species+"/dna/ " + chromosome_fasta_path + " 2>&1>> install.log")

							#os.system("gunzip references/chromosomes/fasta/" + genome + "/*.gz")
							os.system("gzip -d references/chromosomes/fasta/" + genome + "/*.gz")

						except:
							print("The ensembl ftp site is not currently responding.")
							sys.exit()

				print("Chromosome fasta files for " + genome + " have been installed. Creating the chromosome string files now...")
				os.system("python scripts/save_chrom_strings.py -g " + genome)
				print("Chromosome string files for " + genome + " have been created. Continuing with installation.")
				# os.system("rm -r " + chromosome_fasta_path)
				# os.remove(chromosome_fasta_path)
				shutil.rmtree(chromosome_fasta_path)

			else:
				print("Chromosome reference files exist for " + genome + ". Continuing with installation.")


def install_chromosomes_tsb (genomes, ref_dir, custom):

	for genome in genomes:
		chrom_number = None
		if genome == 'GRCh37' or genome == 'GRCh38': 
			chrom_number = 24
		elif genome == 'mm10' or genome == 'mm9':
			chrom_number = 21
		elif genome == 'rn6':
			chrom_number = 22

		if custom:
			chrom_number = len([x for x in os.listdir("references/chromosomes/chrom_string/" + genome + "/") if x != ".DS_Store"])


		chromosome_TSB_path = "references/chromosomes/tsb/" + genome + "/"
		transcript_files = "references/chromosomes/transcripts/" + genome + "/"
		print("[DEBUG] Chromosome tsb files found at: " + ref_dir +  chromosome_TSB_path)


		if os.path.exists(transcript_files) == False or len(os.listdir(transcript_files)) < 1:
			print("Please download the transcript files before proceeding. You can download the files from 'http://www.ensembl.org/biomart/martview'.")
			print("Follow the format presented in the README file:\n\n\tGene stable ID  Transcript stable ID    Chromosome/scaffold name    Strand  Transcript start (bp)   Transcript end (bp)\n\n\n")
			sys.exit()
		if os.path.exists(chromosome_TSB_path) == False or len(os.listdir(chromosome_TSB_path)) < chrom_number:
			print("The transcriptional reference data for " + genome + " has not been saved. Creating these files now")
			os.system("python scripts/save_tsb_192.py -g " + genome)

		if not custom:
			corrupt = False
			for files in os.listdir(chromosome_TSB_path):
				if "proportions" in files:
					continue
				if ".DS_Store" in files:
					continue
				chrom = files.split(".")
				chrom = chrom[0]
				check = md5(chromosome_TSB_path + files)
				if check_sum[genome][chrom] != check:
					corrupt = True
					os.remove(chromosome_TSB_path + files)
					print("[DEBUG] Chromosome " + chrom + " md5sum did not match => reference md5sum: " + str(check_sum[genome][chrom]) + "    new file md5sum: " + str(check))
			if corrupt:
				print("The transcriptional reference data appears to be corrupted. Please reinstall the " + genome + " genome.")
				sys.exit()
			
		print("The transcriptional reference data for " + genome + " has been saved.")

def install_chromosomes_tsb_BED (genomes, ref_dir, custom):
	for genome in genomes:
		if not os.path.exists(ref_dir + "chromosomes/tsb_BED/" + genome + "/") or len(os.listdir(ref_dir + "chromosomes/tsb_BED/" + genome + "/")) < 19:
			is_custom = False
			if custom:
				is_custom = True
			os.system("python scripts/save_chrom_tsb_separate.py -g " + genome + " -c " + str(is_custom))
			print("The TSB BED files for " + genome + " have been saved.")

def benchmark (genome, ref_dir):
	#current_dir = os.path.realpath(__file__)
	#ref_dir = re.sub('\/install.py$', '', current_dir)
	ref_dir = os.path.dirname(os.path.abspath(__file__))
	vcf_path = ref_dir + "/references/vcf_files/" + genome + "_bench/"

	start_time = time.time()
	matGen.SigProfilerMatrixGeneratorFunc(genome + "_bench", genome, vcf_path)
	end_time = time.time()

	original_matrix_96 = ref_dir + "/scripts/Benchmark/" + genome + "_bench_orig_96.txt"
	original_matrix_3072 = ref_dir + "/scripts/Benchmark/" + genome + "_bench_orig_3072.txt"
	new_matrix_96 = vcf_path + "output/SBS/" + genome + "_bench.SBS96.all"
	new_matrix_3072 = vcf_path + "output/SBS/" + genome + "_bench.SBS6144.all"

	#genome = "GRCh37"

	############# Cosine Test ###################################################
	data_orig = pd.read_csv(original_matrix_96, sep='\t', header=0)
	data_new = pd.read_csv(new_matrix_96, sep='\t', header=0)
	count = 0
	range_count = min(len(data_orig.loc[0]), len(data_new.loc[0]))
	for i in range (1, range_count, 1):
		orig_list = list(data_orig[data_orig.columns[i]])
		new_list = list(data_new[data_new.columns[i]])
		cosine_result = (1-spatial.distance.cosine(orig_list,new_list))
		if cosine_result != 1:
			count += 1
	if count != 0:
		print("There seems to be some errors in the newly generated matrix. The installation may not have been successful.")


	data_orig = pd.read_csv(original_matrix_3072, sep='\t', header=0)
	data_new = pd.read_csv(new_matrix_3072, sep='\t', header=0)
	count = 0
	range_count = min(len(data_orig.loc[0]), len(data_new.loc[0]))
	for i in range (1, range_count, 1):
		orig_list = data_orig[data_orig.columns[i]]
		new_list = data_new[data_new.columns[i]]
		cosine_result = (1-spatial.distance.cosine(orig_list,new_list))
		if cosine_result <= 0.85:
			count += 1
	if count != 0:
		print("There seems to be some errors in the newly generated matrix. The installation may not have been successful.")

	end_time = time.time()
	print("Installation was succesful.\nSigProfilerMatrixGenerator took " + str(end_time-start_time) + " seconds to complete.")


def install (genome, custom=False, rsync=False, bash=True, ftp=True, fastaPath=None, transcriptPath=None, exomePath=None, offline_files_path=None):
	if custom or offline_files_path is not None:
		ftp=False
	first_path= os.getcwd()
	ref_dir = os.path.dirname(os.path.abspath(__file__))
	os.chdir(ref_dir)

	if not custom and offline_files_path is None:
		wget_install = shutil.which("wget") is not None
		rsync_install = shutil.which("rsync") is not None
		if not wget_install and not rsync_install:
			print("You do not have wget nor rsync installed. Please install wget or rsync and then reattempt to install your desired genome.")
			sys.exit()
		elif not wget_install and rsync_install:
			print("You do not have wget installed. Please install wget and then reattempt to install your desired genome.")
			sys.exit()
	tar_install = shutil.which("tar") is not None
	if not tar_install:
		print("You do not have tar installed. Please install tar and then reattempt to install your desired genome.")
		sys.exit()


	if os.path.exists("install.log"):
		os.remove("install.log")

	chrom_string_dir = ref_dir + "/references/chromosomes/chrom_string/"
	chrom_fasta_dir = ref_dir + "/references/chromosomes/fasta/"
	chrom_tsb_dir = ref_dir + "/references/chromosomes/tsb/"
	matrix_dir = ref_dir + "/references/matrix/"
	vcf_dir = ref_dir + "/references/vcf_files/"
	bed_dir = ref_dir + "/references/vcf_files/BED/"
	log_dir = "logs/"
	new_dirs = [ref_dir, chrom_string_dir, chrom_fasta_dir, chrom_tsb_dir, matrix_dir, vcf_dir, bed_dir, log_dir]


	for dirs in new_dirs:
		if not os.path.exists(dirs):
			os.makedirs(dirs)


	if custom:
		transcript_files = "references/chromosomes/transcripts/" + genome + "/"
		if os.path.exists(chrom_fasta_dir + genome + "/"):
			shutil.rmtree(chrom_fasta_dir + genome + "/")
		os.makedirs(chrom_fasta_dir + genome + "/")
		fastaFiles = glob.iglob(os.path.join(fastaPath, "*.gz"))
		for file in fastaFiles:
			if os.path.isfile(file):
				shutil.copy(file, chrom_fasta_dir + genome + "/")

		if os.path.exists(ref_dir + "/references/chromosomes/transcripts/" + genome + "/"):
			shutil.rmtree(ref_dir + "/references/chromosomes/transcripts/" + genome + "/")
		os.makedirs(ref_dir + "/references/chromosomes/transcripts/" +  genome + "/")
		shutil.copy(transcriptPath,ref_dir + "/references/chromosomes/transcripts/" +  genome + "/")

		if os.path.exists(ref_dir + "/references/chromosomes/exome/" + genome + "/"):
			shutil.rmtree(ref_dir + "/references/chromosomes/exome/" + genome + "/")
		os.makedirs(ref_dir + "/references/chromosomes/exome/" +  genome + "/")
		if exomePath is not None:
			shutil.copy(exomePath,ref_dir + "/references/chromosomes/exome/" +  genome + "/")


	if ftp:

		chromosome_fasta_path = ref_dir + "/references/chromosomes/tsb/"
		print("Beginning installation. This may take up to 40 minutes to complete.")
		if not rsync:
			try:
				if bash:
					try:
						os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -nd -P ' + chromosome_fasta_path + ' ftp://alexandrovlab-ftp.ucsd.edu/pub/tools/SigProfilerMatrixGenerator/' + genome + '.tar.gz 2>> install.log' + "'")
					except:
						print("The UCSD ftp site is not responding...pulling from sanger ftp now.")
					try:
						os.system("bash -c '" + 'wget -r -l1 -c -nc --no-parent -nd -P ' + chromosome_fasta_path + ' ftp://ngs.sanger.ac.uk/scratch/project/mutographs/SigProf/' + genome + '.tar.gz 2>> install.log' + "'")
					except:
						print("The Sanger ftp site is not responding. Please check your internet connection/try again later.")
				else:
					os.system('wget -r -l1 -c -nc --no-parent -nd -P ' + chromosome_fasta_path + ' ftp://ngs.sanger.ac.uk/scratch/project/mutographs/SigProf/' + genome + '.tar.gz 2>> install.log')
				os.system("tar -xzf " + ref_dir + "/references/chromosomes/tsb/" + genome + ".tar.gz -C " + ref_dir + "/references/chromosomes/tsb/")
				os.remove(ref_dir + "/references/chromosomes/tsb/" + genome + ".tar.gz")
			except:
				print("The ensembl ftp site is not currently responding.")
				sys.exit()
		else:
			print("Direct download for RSYNC is not yet supported")
			sys.exit()
			# try:
			#     if bash:
			#         os.system("bash -c '" + "rsync -av -m --include='*/' rsync://ftp.ngs.sanger.ac.uk/scratch/project/mutographs/SigProf/" + genome + ".tar.gz " + chromosome_fasta_path + " 2>&1>> install.log" + "'")
			#     else:
			#         os.system("rsync -av -m rsync://ftp://ngs.sanger.ac.uk/scratch/project/mutographs/SigProf/" + genome + ".tar.gz " + chromosome_fasta_path + " 2>&1>> install.log")
			#     os.system("tar -xzf " + ref_dir + "/references/chromosomes/tsb/" + genome + ".tar.gz -C " + ref_dir + "/references/chromosomes/tsb/")
			#     os.remove(ref_dir + "/references/chromosomes/tsb/" + genome + ".tar.gz")
			# except:
			#     print("The ensembl ftp site is not currently responding.")
			#     sys.exit()

		chromosome_TSB_path = chromosome_fasta_path + genome + "/"
		corrupt = False
		for files in os.listdir(chromosome_TSB_path):
			if "proportions" in files:
				continue
			if ".DS_Store" in files:
				continue
			chrom = files.split(".")
			chrom = chrom[0]
			check = md5(chromosome_TSB_path + files)
			if check_sum[genome][chrom] != check:
				corrupt = True
				os.remove(chromosome_TSB_path + files)
				print("[DEBUG] Chromosome " + chrom + " md5sum did not match => reference md5sum: " + str(check_sum[genome][chrom]) + "    new file md5sum: " + str(check))
		if corrupt:
			print("The transcriptional reference data appears to be corrupted. Please reinstall the " + genome + " genome.")
			sys.exit()
		print("The transcriptional reference data for " + genome + " has been saved.")

	elif offline_files_path is not None:
		print("Beginning installation using locally provided files.")
		
		# unpack user provided tar file into environment
		cur_dir = os.getcwd()
		os.chdir(first_path)
		shutil.unpack_archive(offline_files_path + genome + ".tar.gz", ref_dir + "/references/chromosomes/tsb/")
		os.chdir(cur_dir)
		
		chromosome_fasta_path = ref_dir + "/references/chromosomes/tsb/"
		chromosome_TSB_path = chromosome_fasta_path + genome + "/"
		corrupt=False
		
		for files in os.listdir(chromosome_TSB_path):
			if "proportions" in files:
				continue
			if ".DS_Store" in files:
				continue
			chrom = files.split(".")
			chrom = chrom[0]
			check = md5(chromosome_TSB_path + files)
			if check_sum[genome][chrom] != check:
				corrupt = True
				os.remove(chromosome_TSB_path + files)
				print("[DEBUG] Chromosome " + chrom + " md5sum did not match => reference md5sum: " + str(check_sum[genome][chrom]) + "    new file md5sum: " + str(check))
		if corrupt:
			print("The transcriptional reference data appears to be corrupted. Please reinstall the " + genome + " genome.")
			sys.exit()
		print("The transcriptional reference data for " + genome + " has been saved.")
		
	else:
		print("Beginning installation. This may take up to 20 minutes to complete.")
		first_path = os.getcwd()
		ref_dir = os.path.dirname(os.path.abspath(__file__))
		os.chdir(ref_dir)

		print("[DEBUG] Path to SigProfilerMatrixGenerator used for the install: ", ref_dir)

		genomes = [genome]

		if os.path.exists("install.log"):
			os.remove("install.log")

		install_chromosomes(genomes, ref_dir, custom, rsync, bash)
		install_chromosomes_tsb (genomes, ref_dir, custom)

		if custom:
			install_chromosomes_tsb_BED (genomes, ref_dir, custom)

	if os.path.exists("BRCA_example/"):
		shutil.copy("BRCA_example/", "references/vcf_files/")
	if os.path.exists("example_test"):
		shutil.copy("example_test/", "references/vcf_files/")
	if os.path.exists("context_distributions/"):
		shutil.copy("context_distributions/", "references/chromosomes/")





	print("All reference files have been created.")
	if 'havana' in genome:
		genome = genome.split("_")[0]
	if genome != "rn6" and genome != 'dog' and genome != 'c_elegans' and genome != 'mm39' and not custom:
		print("Verifying and benchmarking installation now...")
		benchmark(genome, ref_dir)

	print ("To proceed with matrix_generation, please provide the path to your vcf files and an appropriate output path.")
	shutil.rmtree(chrom_string_dir)
	print("Installation complete.")
	os.chdir(first_path)

def main ():

	first_path= os.getcwd()
	os.chdir(first_path + "/sigProfilerMatrixGenerator/")
	genomes = ['mm9', 'mm10','GRCh37', 'GRCh38' ]
	#genomes = ['GRCh37']
	custom = False
	parser = argparse.ArgumentParser(description="Provide the necessary arguments to install the reference files.")
	parser.add_argument("-g", "--genome", nargs='?', help="Optional parameter instructs script to install the custom genome.")
	parser.add_argument("-ct", "--custom", help="Optional parameter instructs script to create the reference files for a custom genome", action='store_true')
	args = parser.parse_args()

	if args.genome:
		genomes = [args.genome]
	if args.custom:
		custom = True

	if os.path.exists("install.log"):
		os.system("rm install.log")

	ref_dir = "references/"
	chrom_string_dir = ref_dir + "chromosomes/chrom_string/"
	chrom_fasta_dir = ref_dir + "chromosomes/fasta/"
	chrom_tsb_dir = ref_dir + "chromosomes/tsb/"
	matrix_dir = ref_dir + "matrix/"
	vcf_dir = ref_dir + "vcf_files/"
	bed_dir = ref_dir + "vcf_files/BED/"
	log_dir = "logs/"
	new_dirs = [ref_dir, chrom_string_dir, chrom_fasta_dir, chrom_tsb_dir, matrix_dir, vcf_dir, bed_dir, log_dir]

	current_dir = os.getcwd()
	for dirs in new_dirs:
		if not os.path.exists(dirs):
			os.makedirs(dirs)


	install_chromosomes(genomes, ref_dir, custom)
	install_chromosomes_tsb (genomes, ref_dir, custom)
	#install_chromosomes_tsb_BED (genomes, custom, ref_dir)
	if os.path.exists("BRCA_example/"):
		os.system("mv BRCA_example/ references/vcf_files/")
	if os.path.exists("example_test"):
		os.system("mv example_test/ references/vcf_files/")
	if os.path.exists("context_distributions/"):
		os.system("mv context_distributions/ references/chromosomes/")

	
	if os.path.exists(chrom_tsb_dir + "GRCh37/"):
		print("All reference files have been created.\nVerifying and benchmarking installation now...")
		benchmark(ref_dir)
	else:
		print("All reference files have been created.")
	print ("Please place your vcf files for each sample into the 'references/vcf_files/[test]/[mutation_type]/' directory. Once you have done that, you can proceed with the matrix generation.")
	#os.system("rm -r " + chrom_string_dir)
	print("Installation complete.")
	os.chdir(first_path)

if __name__ == '__main__':
	main()



import sys
import collections
import itertools
import functools
import warnings

import pysam
import Bio.Align

import importlib
top_package_name = __name__.split(".")[0]
common = importlib.import_module(".".join([top_package_name, "common"]))
libvcfspec = importlib.import_module(".".join([top_package_name, "variant", "vcfspec"]))
libpileup = importlib.import_module(".".join([top_package_name, "read", "pileup"]))
alignhandler = importlib.import_module(".".join([top_package_name, "align", "alignhandler"]))
bameditor = importlib.import_module(".".join([top_package_name, "bameditor"]))
readhandler = importlib.import_module(".".join([top_package_name, "read", "readhandler"]))
fetchcache = importlib.import_module(".".join([top_package_name, "read", "fetchcache"]))


DEFAULT_INACTIVE_PADDING = 10
DEFAULT_ACTIVE_THRESHOLD = 0.05
DEFAULT_VCFSPEC_EXTRACTION_THRESHOLD = 0.05


#ALIGNER_FILLED = Bio.Align.PairwiseAligner(
#    mode='local',
#    match_score=2,
#    mismatch_score=-3,
#    query_internal_open_gap_score=-7,
#    query_internal_extend_gap_score=-2,
#    target_internal_open_gap_score=-7,
#    target_internal_extend_gap_score=-2,
#    query_left_open_gap_score=-7,
#    query_left_extend_gap_score=-2,
#)

#ALIGNER_UNFILLED = Bio.Align.PairwiseAligner(
#    mode='local',
#    match_score=2,
#    mismatch_score=-3,
#    query_internal_open_gap_score=-7,
#    query_internal_extend_gap_score=-2,
#    target_internal_open_gap_score=-7,
#    target_internal_extend_gap_score=-2,
#)

ALIGNER_BLASTN = Bio.Align.PairwiseAligner(
    match_score=2,
    mismatch_score=-3,
    query_internal_open_gap_score=-7,
    query_internal_extend_gap_score=-2,
    target_internal_open_gap_score=-7,
    target_internal_extend_gap_score=-2,
)

ALIGNER_EQUAL_MM_GAP = Bio.Align.PairwiseAligner(
    mode='global',
    match_score=3,
    mismatch_score=-3,
    query_internal_open_gap_score=-3,
    query_internal_extend_gap_score=0,
    target_internal_open_gap_score=-3,
    target_internal_extend_gap_score=0,
)

ALIGNER_MAIN = ALIGNER_EQUAL_MM_GAP



###########################
# secure_inactive_padding #
###########################

def secure_inactive_padding_rightward(pileup, inactive_padding, extend_pileup_by=libpileup.DEFAULT_EXTEND_PILEUP_BY, inplace=False):
    """Returns a modified pileup object with inactive region on the right margin"""
    # set initial_rightmost_active_pos0
    initial_active_positions = pileup.get_active_positions()
    if len(initial_active_positions) == 0:
        initial_rightmost_active_pos0 = pileup.start0 - 1
    else:
        initial_rightmost_active_pos0 = max(initial_active_positions)
    # initial extend
    while True:
        if pileup.end0 - (initial_rightmost_active_pos0 + 1) < inactive_padding:
            pileup.extend_rightward(width=extend_pileup_by)
        else:
            break
    # initial search loop
    active_info = pileup.get_active_info(start0=(initial_rightmost_active_pos0 + 1))
    found_result, farthest_inactive_pos0 = _find_inactive_run(active_info, inactive_padding)
    # subsequent search loop
    if not found_result:
        while True:
            pileup.extend_rightward(width=extend_pileup_by)
            active_info = pileup.get_active_info(
                start0=(pileup.end0 - extend_pileup_by + 1 - inactive_padding)
            )
            found_result, farthest_inactive_pos0 = _find_inactive_run(active_info, inactive_padding)
            if found_result:
                break
    # clip pileup
    if inplace:
        pileup.subset(pileup.start0, farthest_inactive_pos0 + 1, inplace=True)
    else:
        pileup_clipped = pileup.subset(pileup.start0, farthest_inactive_pos0 + 1, inplace=False)
        return pileup_clipped


def secure_inactive_padding_leftward(pileup, inactive_padding, extend_pileup_by=libpileup.DEFAULT_EXTEND_PILEUP_BY, inplace=False):
    """Returns a modified pileup object with inactive region on the left margin"""
    # set initial_leftmost_active_pos0
    initial_active_positions = pileup.get_active_positions()
    if len(initial_active_positions) == 0:
        initial_leftmost_active_pos0 = pileup.end0
    else:
        initial_leftmost_active_pos0 = min(initial_active_positions)
    # initial extend
    while True:
        if initial_leftmost_active_pos0 - pileup.start0 < inactive_padding:
            pileup.extend_leftward(width=extend_pileup_by)
        else:
            break
    # initial search loop
    active_info = pileup.get_active_info(end0=initial_leftmost_active_pos0)[::-1]
    found_result, farthest_pos0 = _find_inactive_run(active_info, inactive_padding)
    # subsequent search loop
    if not found_result:
        while True:
            pileup.extend_leftward(width=extend_pileup_by)
            active_info = pileup.get_active_info(
                end0=(pileup.start0 + extend_pileup_by - 1 + inactive_padding)
            )[::-1]
            found_result, farthest_pos0 = _find_inactive_run(active_info, inactive_padding)
            if found_result:
                break
    # clip pileup
    if inplace:
        pileup.subset(farthest_pos0, pileup.end0, inplace=True)
    else:
        pileup_clipped = pileup.subset(farthest_pos0, pileup.end0, inplace=False)
        return pileup_clipped


def _find_inactive_run(active_info, inactive_padding):
    found_result = False
    farthest_inactive_pos0 = None
    for start_idx in range(0, len(active_info) - inactive_padding + 1):
        if not any(active_info[start_idx:(start_idx + inactive_padding)]):
            found_result = True
            farthest_inactive_pos0 = active_info.index[start_idx + inactive_padding -1]
            break
    return found_result, farthest_inactive_pos0


###########################
# pileup range correction #
###########################

def augment_pileup_range(active_region_pileup, factor_vcfspec_range=2, factor_repeat_area=1.5):
    active_region_pileup.set_row_specs()
    active_region_pileup.set_row_spec_groups()
    pileup_vcfspecs = get_vcfspecs_from_pileup(active_region_pileup)

    candidate_start0s = list()
    candidate_end0s = list()
    for vcfspec in itertools.chain.from_iterable(pileup_vcfspecs):
        # padding with (vcfspec REF_range length) * factor
        vcfspec_range = vcfspec.REF_range0
        spanning_length = len(vcfspec_range)
        padding = int(len(vcfspec_range) * factor_vcfspec_range)
        candidate_start0s.append(vcfspec_range.start - padding)
        candidate_end0s.append(vcfspec_range.stop + padding)
        # padding considering repetitive reference region
        equivalents = vcfspec.get_equivalents(active_region_pileup.fasta)


############################
# get_active_region_pileup #
############################

def get_active_region_pileup(
    chrom,
    start0,
    end0,
    bam,
    fasta,
    active_threshold=DEFAULT_ACTIVE_THRESHOLD,
    inactive_padding=DEFAULT_INACTIVE_PADDING,
    fetchedreads_addition_padding=fetchcache.DEFAULT_FETCHEDREADS_ADDITION_PADDING,
    extend_pileup_by=libpileup.DEFAULT_EXTEND_PILEUP_BY,
    set_row_specs=True,
):
    """Returns:
        "active_range" and "pileup" are returned.
        When there is no active position around input range (start0, end0), 
            returned "active_range" is None. Pileup object, created as a 
            byproduct, is returned.
        Otherwise, "active_range" is python range object. Pileup object is
            clipped by "subset" method to fit "active_range", 
            then "set_row_specs" and "set_row_spec_groups" methods are run, 
            then returned.
    """

    def get_initial_fetch_range_old(start0, end0, extend_fetchedreads_by):
        query_range = range(start0, end0)
        diff_range_len = extend_fetchedreads_by - len(query_range)
        if diff_range_len > 0:
            pad = int(diff_range_len / 2)
            if diff_range_len % 2 == 0:
                initial_fetch_range = range(start0 - pad, end0 + pad)
            else:
                initial_fetch_range = range(start0 - pad, end0 + pad + 1)
        else:
            initial_fetch_range = query_range

        return initial_fetch_range

    def get_initial_fetch_range(start0, end0, inactive_padding):
        return range(start0 - inactive_padding * 3, end0 + inactive_padding * 3)

    def inspect_initial_range(
        chrom, start0, end0, fasta, bam, active_threshold, inactive_padding, fetchedreads_addition_padding,
    ):
        initial_fetch_range = get_initial_fetch_range(start0, end0, inactive_padding)
        fetchedreads_initial = fetchcache.FetchedReads.from_fetch(
            bam, chrom, initial_fetch_range.start, initial_fetch_range.stop, 
            readfilter=readhandler.readfilter_pileup, 
            addition_padding=fetchedreads_addition_padding,
        )
        active_region_pileup = libpileup.get_pileup(
            chrom=chrom, 
            start0=start0, 
            end0=end0, 
            fetchedreads=fetchedreads_initial,
            fasta=fasta,
            active_threshold=active_threshold,
            truncate=True,
            as_array=False,
            return_range=False,
        )
        return active_region_pileup

#    def augment_inactive_padding(active_region_pileup, inactive_padding, extend_pileup_by, extend_fetchedreads_by, factor=2):
#        active_region_length = len(active_region_pileup.range0) - (2 * inactive_padding)
#        if inactive_padding < factor * active_region_length:
#            new_inactive_padding = factor * active_region_length
#            active_region_pileup.secure_inactive_padding_rightward(new_inactive_padding, extend_pileup_by, extend_fetchedreads_by)
#            active_region_pileup.secure_inactive_padding_leftward(new_inactive_padding, extend_pileup_by, extend_fetchedreads_by)

    def extend_pileup_by_active_area_length(active_region_pileup, inactive_padding, extend_pileup_by, extend_fetchedreads_by, factor=1.3):
        active_area_end0 = active_region_pileup.end0 - inactive_padding
        active_area_start0 = active_region_pileup.start0 + inactive_padding
        active_area_length = active_area_end0 - active_area_start0
        # correct right margin
        desired_pileup_end0 = active_area_end0 + int(active_area_length * factor)
        if active_region_pileup.end0 < desired_pileup_end0:
            active_region_pileup.extend_rightward(width=(desired_pileup_end0 - active_region_pileup.end0), extend_fetchedreads_by=extend_fetchedreads_by)
        # correct left margin
        desired_pileup_start0 = active_area_start0 - int(active_area_length * factor)
        if desired_pileup_start0 < active_region_pileup.start0:
            active_region_pileup.extend_leftward(width=(active_region_pileup.start0 - desired_pileup_start0), extend_fetchedreads_by=extend_fetchedreads_by)
        # secure inactive margin
        active_region_pileup.secure_inactive_padding(inactive_padding, extend_pileup_by, extend_fetchedreads_by)

    def extend_pileup_by_reference_repeat(active_region_pileup, extend_pileup_by, extend_fetchedreads_by, factor=1.3):
        pass

    # main
    active_region_pileup = inspect_initial_range(
        chrom, start0, end0, fasta, bam, active_threshold, inactive_padding, fetchedreads_addition_padding,
    )
    active_region_pileup.secure_inactive_padding(inactive_padding, extend_pileup_by)

    if set_row_specs:
        active_region_pileup.set_row_specs()
        active_region_pileup.set_row_spec_groups()

    active_poslist = active_region_pileup.get_active_positions()
    if len(active_poslist) == 0:
        active_range = None
    else:
        active_range = active_region_pileup.range0

    return active_range, active_region_pileup


def get_active_region_pileup_multisample(
    chrom,
    start0,
    end0,
    bam_dict,
    fasta,
    active_threshold=DEFAULT_ACTIVE_THRESHOLD,
    inactive_padding=DEFAULT_INACTIVE_PADDING,
    fetchedreads_addition_padding=fetchcache.DEFAULT_FETCHEDREADS_ADDITION_PADDING,
    extend_pileup_by=libpileup.DEFAULT_EXTEND_PILEUP_BY,
):
    # initialize pileup_dict
    pileup_dict = dict()
    for sampleid, bam in bam_dict.items():
        active_range, active_region_pileup = get_active_region_pileup(
            chrom,
            start0,
            end0,
            bam,
            fasta,
            active_threshold=active_threshold,
            inactive_padding=inactive_padding,
            fetchedreads_addition_padding=fetchedreads_addition_padding,
            extend_pileup_by=extend_pileup_by,
            set_row_specs=False,
        )
        pileup_dict[sampleid] = active_region_pileup

    # create MultisamplePileup object and postprocess
    active_region_mspileup = libpileup.MultisamplePileup(pileup_dict)
    active_region_mspileup.secure_inactive_padding(inactive_padding, extend_pileup_by=extend_pileup_by)
    active_region_mspileup.set_df()
    active_region_mspileup.set_row_specs()
    active_region_mspileup.set_row_spec_groups()
    active_region_mspileup.divide_row_spec_groups()

    # result values
    active_poslist = active_region_mspileup.get_active_positions()
    if len(active_poslist) == 0:
        active_range = None
    else:
        active_range = active_region_mspileup.range0

    return active_range, active_region_mspileup


####################


def get_realigned_reads(active_region_pileup, active_range, ref_seq, ref_seq_reversed, aligner):
    realigned_reads = dict()
    for superseq_rowid, groupinfo in sorted(
        active_region_pileup.row_spec_groups.items(),
        key=(lambda x: x[1]['subseq_hits']), 
        reverse=True,
    ):
        realigned_read_super, superseq_aln = realign_superseq(superseq_rowid, active_region_pileup, active_range, ref_seq, ref_seq_reversed, aligner)
        realigned_reads[superseq_rowid] = realigned_read_super
        active_region_pileup._alignment_cache[superseq_rowid] = superseq_aln

        superseq_row_spec = active_region_pileup.row_specs[superseq_rowid]
        superseq = superseq_row_spec['seq']
        for subseq_rowid in groupinfo['subseq_rowids']:
            if subseq_rowid in realigned_reads.keys():
                continue
            realigned_read_sub, subseq_aln = realign_subseq(subseq_rowid, active_region_pileup, active_range, superseq, superseq_aln)
            realigned_reads[subseq_rowid] = realigned_read_sub
            
    return realigned_reads


def get_vcfspecs_from_pileup(active_region_pileup, allele_portion_threshold=DEFAULT_VCFSPEC_EXTRACTION_THRESHOLD):
    """Must be run after row_spec_groups is set"""
    # prepare alignments for superseq rows
    unprepared_superseq_rowids = set(active_region_pileup.row_spec_groups.keys()).difference(
        active_region_pileup._alignment_cache.keys()
    )
    active_region_pileup.save_alignments(unprepared_superseq_rowids)
    # main
    vcfspecs = list()
    subseq_hits_sum = sum(x['subseq_hits'] for x in active_region_pileup.row_spec_groups.values())
    for superseq_rowid, groupinfo in active_region_pileup.row_spec_groups.items():
        allele_portion = groupinfo['subseq_hits'] / subseq_hits_sum
        if allele_portion < allele_portion_threshold:
            continue

        superseq_row_spec = active_region_pileup.row_specs[superseq_rowid]
        lstrip_query_gaps = not superseq_row_spec["left_filled"]
        rstrip_query_gaps = not superseq_row_spec["right_filled"]
        superseq_aln = active_region_pileup._alignment_cache[superseq_rowid]

        superseq_vcfspec_tuple = alignhandler.alignment_to_vcfspec(
            alignment=superseq_aln, 
            target_start0=active_region_pileup.start0, 
            chrom=active_region_pileup.chrom, 
            fasta=active_region_pileup.fasta,
            lstrip_query_gaps=lstrip_query_gaps,
            rstrip_query_gaps=rstrip_query_gaps,
        )
        if len(superseq_vcfspec_tuple) > 0:
            vcfspecs.append(superseq_vcfspec_tuple)
            
    return vcfspecs


def get_vcfspecs_multisample_pileup(active_region_mspileup, allele_portion_threshold):
    """Must be run after 'active_region_mspileup' has executed 
    'divide_row_spec_groups' and 'get_realigned_reads' methods.
        'divide_row_spec_groups' is needed for setting 'row_spec_groups' for
            each single sample Pileup.
        'get_realigned_reads' is needed for initiating '_alignment_cache' attribute.
    """
    vcfspecs_by_sample = dict()
    calculated_vcfspecs_cache = dict()
    #{sampleid: list() for sampleid in active_region_mspileup.pileup_dict.keys()}
    for sampleid, pileup in active_region_mspileup.pileup_dict.items():
        vcfspecs_onesample = set()
        subseq_hits_sum = sum(x['subseq_hits'] for x in pileup.row_spec_groups.values())

        for superseq_rowid, groupinfo in pileup.row_spec_groups.items():
            allele_portion = groupinfo['subseq_hits'] / subseq_hits_sum
            if allele_portion < allele_portion_threshold:
                continue

            if superseq_rowid in calculated_vcfspecs_cache.keys():
                superseq_vcfspec_tuple = calculated_vcfspecs_cache[superseq_rowid]
            else:
                superseq_row_spec = active_region_mspileup.row_specs[superseq_rowid]
                lstrip_query_gaps = not superseq_row_spec["left_filled"]
                rstrip_query_gaps = not superseq_row_spec["right_filled"]
                superseq_aln = active_region_mspileup._alignment_cache[superseq_rowid]
                superseq_vcfspec_tuple = alignhandler.alignment_to_vcfspec(
                    alignment=superseq_aln, 
                    target_start0=active_region_mspileup.start0, 
                    chrom=active_region_mspileup.chrom, 
                    fasta=active_region_mspileup.fasta,
                    lstrip_query_gaps=lstrip_query_gaps,
                    rstrip_query_gaps=rstrip_query_gaps,
                )
                calculated_vcfspecs_cache[superseq_rowid] = superseq_vcfspec_tuple

            if len(superseq_vcfspec_tuple) > 0:
                vcfspecs_onesample.add(superseq_vcfspec_tuple)

        vcfspecs_by_sample[sampleid] = vcfspecs_onesample
    # final result
    vcfspecs_allsamples = set(itertools.chain.from_iterable(vcfspecs_by_sample.values()))
            
    return vcfspecs_allsamples


#def get_vcfspec_list_old(active_region_pileup, active_range, ref_seq, allele_portion_threshold=0.05):
#    # step 1: get selected_superseq_uids
#    # selected_superseq_uids = list()
#    vcfspec_list = list()
#    subseq_hits_sum = sum(x['subseq_hits'] for x in active_region_pileup.row_spec_groups.values())
#    for superseq_uid, dic in active_region_pileup.row_spec_groups.items():
#        if dic['subseq_hits'] / subseq_hits_sum > allele_portion_threshold:
#            row_spec = active_region_pileup.row_specs[superseq_uid]
#            if ref_seq == row_spec['seq']:
#                continue
#            vcfspec = libvcfspec.Vcfspec(active_region_pileup.chrom, active_range.start + 1, ref_seq, (row_spec['seq'],))
#            vcfspec = vcfspec.normalize(active_region_pileup.fasta)
#            vcfspec_list.append(vcfspec)
#            
#    return vcfspec_list


def write_realigned_bam(bam_path, realigned_reads, active_region_pileup, padding_width=0):
    """Args:
        realigned_reads: dict (keys: ReadUID, values: pysam.AlignedSegment)
    """
    written_reads_range = range(
        active_region_pileup.start0 - padding_width, 
        active_region_pileup.end0 + padding_width, 
    )
    # write reads
    hdr = pysam.AlignmentHeader.from_references(
        reference_names=active_region_pileup.fasta.references,
        reference_lengths=active_region_pileup.fasta.lengths,
    )
    with pysam.AlignmentFile(bam_path, mode='wb', header=hdr) as in_bam:
        # write non-realigned reads
        for uid, read in active_region_pileup.fetchedreads.fetch(
            active_region_pileup.chrom,
            written_reads_range.start,
            written_reads_range.stop,
            with_uid=True,
        ):
            if uid not in realigned_reads.keys():
                in_bam.write(read)
        # write realigned reads
        for read in realigned_reads.values():
            in_bam.write(read)
    # sort and index
    bameditor.sort_and_index(bam_path)


def write_realigned_bam_multisample(bam_paths, realigned_reads_multisample, active_region_pileup_multisample, padding_width=0):
    for sampleid, path in bam_paths.items():
        write_realigned_bam(
            path,
            realigned_reads_multisample[sampleid],
            active_region_pileup_multisample.pileup_dict[sampleid],
            padding_width=padding_width,
        )


###################################
# helpers for get_realigned_reads #
###################################

#def realign_superseq_old(uid, active_region_pileup, active_range, ref_seq, ref_seq_reversed):
#    read = active_region_pileup.get_read(uid)
#    row_spec = active_region_pileup.row_specs[uid]
#    active_region_alignment = align_row_spec(row_spec, ref_seq, ref_seq_reversed)
#    realigned_read = active_aln_to_full_read(active_region_alignment, read, active_range)
#    return realigned_read, active_region_alignment


def get_alignment_from_rowid(row_id, active_region_pileup, ref_seq, ref_seq_reversed, aligner):
    row_spec = active_region_pileup.row_specs[row_id]
    return align_row_spec(row_spec, ref_seq, ref_seq_reversed, aligner)


def realign_superseq(row_id, active_region_pileup, active_range, ref_seq, ref_seq_reversed, aligner):
    # set params
    original_read = active_region_pileup.get_read(row_id)
    original_cigar_split = alignhandler.split_cigar(original_read.cigartuples, original_read.reference_start, active_range)
    row_spec = active_region_pileup.row_specs[row_id]
    empty_before_active_region = (len(original_cigar_split[0]) == 0)
    empty_after_active_region = (len(original_cigar_split[2]) == 0)
    # do alignment
    active_region_alignment = align_row_spec(row_spec, ref_seq, ref_seq_reversed, aligner)
    # get active region cigartuples
    active_region_cigartuples, active_region_offset = alignhandler.alignment_to_cigartuples(
        active_region_alignment,
        match_as_78=False, del_as_skip=False, 
        left_ins_as_clip=empty_before_active_region, right_ins_as_clip=empty_after_active_region,
        remove_left_del=empty_before_active_region, remove_right_del=empty_after_active_region,
    )
    # split cigartuples are merged into one
    realigned_cigartuples = merge_cigartuples_list(
        [original_cigar_split[0], active_region_cigartuples, original_cigar_split[2]]
    )
    # get new read reference_start
    if empty_before_active_region:
        new_reference_start = active_range.start + active_region_offset
    else:
        new_reference_start = original_read.reference_start
        
    realigned_read = original_read.__copy__()
    realigned_read.reference_start = new_reference_start
    realigned_read.cigartuples = realigned_cigartuples
    readhandler.set_NMMD(realigned_read, active_region_pileup.fasta)

    realigned_cigar_sanity_check(original_read, realigned_read, realigned_cigartuples)

    return realigned_read, active_region_alignment


#def realign_subseq_old(uid, active_region_pileup, active_range, superseq, superseq_reversed, superseq_alignment):
#    # set params
#    read = active_region_pileup.get_read(uid)
#    row_spec = active_region_pileup.row_specs[uid]
#        
#    sub_to_sup_aln = align_row_spec(row_spec, superseq, superseq_reversed)
#
#    # sanity check
#    active_region_cigartuples, active_region_offset = alignhandler.alignment_to_cigartuples(sub_to_sup_aln)
#    if set(x[0] for x in active_region_cigartuples) != alignhandler.CIGAROPS_BOTH:
#        raise Exception(
#            f'subseq-to-superseq alignment is not match-only.\n'
#            f'subseq row_spec: {row_spec}\n'
#            f'subseq-to-superseq alignment: {sub_to_sup_aln}'
#        )
#    
#    active_region_alignment = alignhandler.map_alignments(superseq_alignment, sub_to_sup_aln)   
#    # superseq_alignment.map(sub_to_sup_aln)
#    
#    realigned_read = active_aln_to_full_read(active_region_alignment, read, active_range)
#    return realigned_read, active_region_alignment


def realign_subseq(row_id, active_region_pileup, active_range, superseq, superseq_alignment):
    # set params
    original_read = active_region_pileup.get_read(row_id)
    original_cigar_split = alignhandler.split_cigar(original_read.cigartuples, original_read.reference_start, active_range)
    row_spec = active_region_pileup.row_specs[row_id]
    empty_before_active_region = (len(original_cigar_split[0]) == 0)
    empty_after_active_region = (len(original_cigar_split[2]) == 0)

    # do initial alignment
    active_region_alignment = align_subseq_to_ref(row_spec, superseq_alignment)

    # get active region cigartuples
    active_region_cigartuples, active_region_offset = alignhandler.alignment_to_cigartuples(
        active_region_alignment,
        match_as_78=False, del_as_skip=False, 
        left_ins_as_clip=empty_before_active_region, right_ins_as_clip=empty_after_active_region,
        remove_left_del=empty_before_active_region, remove_right_del=empty_after_active_region,
    )
    # merge split cigartuples
    realigned_cigartuples = merge_cigartuples_list(
        [original_cigar_split[0], active_region_cigartuples, original_cigar_split[2]]
    )
    # get new read reference_start
    if empty_before_active_region:
        new_reference_start = active_range.start + active_region_offset
    else:
        new_reference_start = original_read.reference_start
        
    realigned_read = original_read.__copy__()
    realigned_read.reference_start = new_reference_start
    realigned_read.cigartuples = realigned_cigartuples
    readhandler.set_NMMD(realigned_read, active_region_pileup.fasta)

    realigned_cigar_sanity_check(original_read, realigned_read, realigned_cigartuples)

    return realigned_read, active_region_alignment


def align_subseq_to_ref(subseq_row_spec, sup_to_ref_aln):
    # get subseq index
    reverse_align = (not subseq_row_spec['left_filled']) and subseq_row_spec['right_filled']
    if reverse_align:
        sub_to_sup_index = sup_to_ref_aln.query.rfind(subseq_row_spec['seq'])
    else:
        sub_to_sup_index = sup_to_ref_aln.query.find(subseq_row_spec['seq'])
    if sub_to_sup_index == -1:
        raise Exception(
            f'"subseq" is not a subseq of "superseq":\n'
            f'subseq_row_spec: {subseq_row_spec}\n'
            f'superseq: {sup_to_ref_aln.query}'
        )
    # convert into sub-to-ref alignment
    # supaln: sup_to_ref_aln
    # subaln: sub_to_sup_aln
    active_region_walks = list()
    target_walk_subaln = range(sub_to_sup_index, sub_to_sup_index + len(subseq_row_spec['seq']))
    query_walk_subaln = range(0, len(subseq_row_spec['seq']))
    found_initial_hit = False
    for target_walk_supaln, query_walk_supaln in alignhandler.get_walks(sup_to_ref_aln, copy=False):
        if not found_initial_hit:
            if target_walk_subaln.start in query_walk_supaln:
                found_initial_hit = True
                # superseq walk
                superseq_walk_start = target_walk_subaln.start
                superseq_walk_stop = min(query_walk_supaln.stop, target_walk_subaln.stop)
                len_current_superseq_walk = superseq_walk_stop - superseq_walk_start
                # subseq walk
                subseq_walk = range(0, len_current_superseq_walk)
                # ref walk
                if len(target_walk_supaln) == 0:
                    ref_walk = target_walk_supaln
                else:
                    # During initial hit query_walk_supaln cannot be zero-length
                    ref_walk_start = target_walk_supaln.start + (superseq_walk_start - query_walk_supaln.start)
                    ref_walk = range(ref_walk_start, ref_walk_start + len_current_superseq_walk)

                active_region_walks.append((ref_walk, subseq_walk))

                if query_walk_supaln.stop >= target_walk_subaln.stop:
                    break
        else:
            # superseq walk
            superseq_walk_start = query_walk_supaln.start
            superseq_walk_stop = min(query_walk_supaln.stop, target_walk_subaln.stop)
            len_current_superseq_walk = superseq_walk_stop - superseq_walk_start
            # subseq walk
            subseq_walk_start = active_region_walks[-1][1].stop
            subseq_walk = range(subseq_walk_start, subseq_walk_start + len_current_superseq_walk)
            # ref walk
            if len(target_walk_supaln) == 0 or len(query_walk_supaln) == 0:
                ref_walk = target_walk_supaln
            else:
                ref_walk_start = target_walk_supaln.start
                ref_walk = range(ref_walk_start, ref_walk_start + len_current_superseq_walk)

            active_region_walks.append((ref_walk, subseq_walk))

            if query_walk_supaln.stop >= target_walk_subaln.stop:
                break
    # pad with gap: right side
    last_ref_walk = active_region_walks[-1][0]
    last_subseq_walk = active_region_walks[-1][1]
    if last_ref_walk.stop < len(sup_to_ref_aln.target):
        added_ref_walk = range(last_ref_walk.stop, len(sup_to_ref_aln.target))
        added_subseq_walk = range(last_subseq_walk.stop, last_subseq_walk.stop)
        active_region_walks.append((added_ref_walk, added_subseq_walk))
    # pad with gap: left
    first_ref_walk = active_region_walks[0][0]
    first_subseq_walk = active_region_walks[0][1]
    if first_ref_walk.start > 0:
        added_ref_walk = range(0, first_ref_walk.start)
        added_subseq_walk = range(0, 0)
        active_region_walks.insert(0, (added_ref_walk, added_subseq_walk))
    # make result alignment
    return Bio.Align.PairwiseAlignment(
        sup_to_ref_aln.target,
        subseq_row_spec['seq'],
        alignhandler.walks_to_path(active_region_walks),
        0
    )


#######################################
# amenders: marginal insdel amendment #
#######################################

def amender_left_right(alignment):
    return alignhandler.amend_outer_insdel_right(
        alignhandler.amend_outer_insdel_left(alignment)
    )


def amender_left(alignment):
    return alignhandler.amend_outer_insdel_left(alignment)


def amender_right(alignment):
    return alignhandler.amend_outer_insdel_right(alignment)


def amender_none(alignment):
    return alignment


def get_amender(empty_before_active_region, empty_after_active_region):
    if empty_before_active_region:
        if empty_after_active_region:
            return amender_left_right
        else:
            return amender_left
    else:
        if empty_after_active_region:
            return amender_right
        else:
            return amender_none


#######################################


def align_row_spec(row_spec, target, target_reversed, aligner):
    # set params
    query = row_spec['seq']
    reverse_align = (not row_spec['left_filled']) and row_spec['right_filled']
    # run aligner
    try:
        if reverse_align:
            alns = aligner.align(target_reversed, query[::-1])
        else:
            alns = aligner.align(target, query)
    except Exception as exc:
        raise Exception(f'Failed alignment:\nrow_spec: {row_spec}\ntarget: {target}') from exc
    # recover reversed alignment
    if reverse_align:
        alns = [alignhandler.reverse_alignment(x) for x in alns]

    if len(alns) == 1:
        aln = alns[0]
        aln = amender_left_right(aln)
    else:
        alns = [amender_left_right(x) for x in alns]
        try:
            aln = alignhandler.alignment_tiebreaker(alns, raise_with_failure=True)
        except alignhandler.AlignmentTieError as exc:
            msg = f'Failed to break alignment tie. row_spec is:\n{row_spec}'
            raise Exception(msg) from exc
        
    return aln


def align_row_spec_old(row_spec, target, target_reversed, aligner, empty_before_active_region, empty_after_active_region):
    # set params
    query = row_spec['seq']
    reverse_align = (not row_spec['left_filled']) and row_spec['right_filled']
    amender = get_amender(empty_before_active_region, empty_after_active_region)
    # run aligner
    try:
        if reverse_align:
            alns = aligner.align(target_reversed, query[::-1])
        else:
            alns = aligner.align(target, query)
    except Exception as exc:
        raise Exception(f'Failed alignment:\nrow_spec: {row_spec}\ntarget: {target}') from exc
    # recover reversed alignment
    if reverse_align:
        alns = [alignhandler.reverse_alignment(x) for x in alns]

    if len(alns) == 1:
        aln = alns[0]
        aln = amender(aln)
    else:
        alns = [amender(x) for x in alns]
        try:
            aln = alignhandler.alignment_tiebreaker(alns, raise_with_failure=True)
        except alignhandler.AlignmentTieError as exc:
            msg = f'Failed to break alignment tie. row_spec is:\n{row_spec}'
            raise Exception(msg) from exc
        
    return aln


#def active_aln_to_full_read(active_region_alignment, read, active_range):
#    original_cigar_split = alignhandler.split_cigar(read.cigartuples, read.reference_start, active_range)
#
#    active_region_cigartuples, active_region_offset = alignhandler.alignment_to_cigartuples(active_region_alignment)
#    
#    realigned_cigartuples = merge_cigartuples_list(
#        [original_cigar_split[0], active_region_cigartuples, original_cigar_split[2]]
#    )
#    
#    if len(original_cigar_split[0]) > 0:
#        #assert active_region_offset == 0
#        new_reference_start = read.reference_start
#    else:
#        new_reference_start = active_range.start + active_region_offset
#        
#    realigned_read = read.__copy__()
#    realigned_read.reference_start = new_reference_start
#    realigned_read.cigartuples = realigned_cigartuples
#
#    # sanity check
#    # if len(read.query_sequence) != alignhandler.get_query_length(realigned_cigartuples):
#    #     raise Exception(
#    #         f'Read query sequence length and cigar length differs:\n'
#    #         f'Original read: {read.to_string()}\n'
#    #         f'Realigned read: {realigned_read.to_string()}\n'
#    #     )
#
#    return realigned_read


def merge_two_cigartuples(cigartuples_left, cigartuples_right):
    if len(cigartuples_left) == 0 or len(cigartuples_right) == 0:
        return cigartuples_left + cigartuples_right
    else:
        if cigartuples_left[-1][0] == cigartuples_right[0][0]:
            return (
                cigartuples_left[:-1]
                + [(cigartuples_left[-1][0], cigartuples_left[-1][1] + cigartuples_right[0][1])]
                + cigartuples_right[1:]
            )
        else:
            return cigartuples_left + cigartuples_right

    
def merge_cigartuples_list(cigartuples_list):
    return list(functools.reduce(merge_two_cigartuples, cigartuples_list))


def realigned_cigar_sanity_check(original_read, realigned_read, realigned_cigartuples):
    if len(original_read.query_sequence) != alignhandler.get_query_length(realigned_cigartuples):
        raise Exception(
            f'Read query sequence length and cigar length differs.\n'
            f'Original read:\n{original_read.to_string()}\n'
            f'Realigned read:\n{realigned_read.to_string()}\n'
        )

    
###################################
